/*->c.obj */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <assert.h>

#include "os.h"
#include "wimpt.h"

#include "etc.h"
#include "flex.h"
#include "alloc.h"
#include "crc.h"
#include "bits.h"

#include "fsx.h"

#include "bf.h"
#include "err.h"
#include "deb.h"

#include "main.h"
#include "reslink.h"

#include "obj.h"




#ifdef MEMTEST

#define salloc(a,b) flex_alloc(a,b)
#define sfree(a) flex_free(a)

#endif


typedef struct objstr
{
 unsigned int magic:14;
 signed   int balance:2;
 unsigned int copyclass:2;
 unsigned int flags:14;

 int  sibls[2];    /* left/right child nodes */
 int  parent;
 int  free;        /* free space in this object and in child nodes */
 int  count;

 int  length;      /* total length of object */


 int  key;         /* key for searches       */
 int  type;
 char data[1];

} objstr;          /* size 28+3*4  */


#define OVERHEAD    (sizeof(objstr)-3*sizeof(int))
#define MINSIZE     (sizeof(objstr)-1*sizeof(int))


#define FIRSTOBJECT 0xFF
#define DMCHUNK     0x1000
#define DMMAGIC     0xAAA

#define MIRROR      0x1


/* for instance objects, count is number of copies */




#define NIL_NODE     -1    /* ptr for empty (sub)trees */

typedef int DIRECT;          /* type for indexing siblings */

#define        L       ((DIRECT) 0)
#define        R       ((DIRECT) 1)

#define        LEFT    sibls[L]       /* left sibling pointer NODE field */
#define        RIGHT   sibls[R]       /* right sibling pointer NODE field */


/*
 * Direction gives direction in which child NODE c is under parent NODE p.
 */

#define        direction(p,c) ((DIRECT) (((char*)(p))+(p)->RIGHT==(char*)(c)))

/*
 * Cmp_dir gives direction corresponding with compare value v (R iff v > 0).
 */

#define        cmp_dir(v)      ((DIRECT) ((v) > 0))

/*
 * Siblingp yields ptr to left (dr == L) or right (dr == R) child of NODE n.
 */


#define        siblingp(ptr,dir) ((objstr*) (((ptr)->sibls[dir])+(char*)(ptr)))
#define        sibling(ptr, dr)  ((ptr)->sibls + (dr))


/*
 * Parentp yields ptr to parent's ptr to NODE n, or root ptr r if n is 0.
 */


/*

#define        parentp(r, n)   ((xptr((n))->parent) == NIL_NODE ? (r) : \
               siblingp((xptr((n))->parent), direction((xptr((n))->parent), (n))))

 */



/*
 * Dir_bal yields balance value corresponding to DIRECT d.
 */

#define        dir_bal(d)      ((d) == L ? -1 : 1)


/* return pointer to object with offset p in domain d */

#define        objptr(p,d) (((objstr*)((d)->domaindata+(p))))



static os_error * ob_compactallx(domainstr * x);
static objstr * tfind(int key,domainstr * domain);
static void lockblock(objstr * objlo,objstr * objhi,int shift,domainstr * d);



static flexblockstr   ofbb;
static flexblockstr * ofb;


static domainstr * firstdomain;


#ifdef _DEBUG

static void xob_test(domainstr * d,char * where)
{
	os_error * err;

 err=ob_test(d);
 if(err) cprintf("ob test reports %s @ %s",err->errmess,where);

}

#endif


static os_error * dextend(domainstr * d,int newsize)
{
 os_error * err=NULL;
 int        i;

 for(i=0;i<2;i++)
 {
  err=flex_extenda((flex_ptr)&d->domaindata,newsize,ofb);
  if(!err) break;
  ob_compactallx(d);
 }

 return(err);
}



static os_error * dchunk(domainstr * d,int step,objstr ** obj)
{
 os_error * err=NULL;
 int        i;
 char     * olddata;  
 int        shift;

 
 for(i=0;i<2;i++)
 {
  olddata=d->domaindata;
  err=flex_chunka((flex_ptr)&d->domaindata,d->size+step,DMCHUNK,ofb);
  if(!err)
  {
   shift=d->domaindata-olddata;
   if(shift)
   {
    objstr   * objlo;
    objstr   * objhi;

    objlo=(objstr*)(d->domaindata+d->bottom-shift);
    objhi=(objstr*)(((char*)tfind(d->maxobject,d))-shift);
    lockblock(objlo,objhi,shift,d);
    
    if(obj) *obj=(objstr*)(((char*)(*obj))+shift);
   }

   break;
  }
  ob_compactallx(d);
 }

 return(err);
}



static os_error * dalloc(domainstr * d,int size)
{
 os_error * err=NULL;
 int        i;

 for(i=0;i<2;i++)
 {
  err=flex_alloca((flex_ptr)&d->domaindata,size,ofb);
  if(!err) break;
  ob_compactallx(d);
 }

 return(err);
}



static os_error * dfree(domainstr * d)
{
 os_error * err;

 if(d->domaindata) err=flex_freea((flex_ptr)&d->domaindata,ofb);
 else              err=NULL;

 return(err);
}







/*
 * Parentp yields ptr to parent's ptr to NODE n, or root ptr r if n is 0.
 */


static int * parentp(objstr * sb,domainstr * domain)
{
 int    * q;
 objstr * pa;

 if(sb->parent==NIL_NODE) return(&domain->root);

 pa=(objstr*)((char*)sb+sb->parent);

 q=sibling(pa,direction(pa,sb));

 return(q);
}



/* how much free space is there in an object */

static int tfreenode(objstr * sb)
{
 objstr * child;
 int      free;

 free=sb->free;
 if(sb->LEFT!=NIL_NODE)
 {
  child=(objstr*)((char*)sb+sb->LEFT);
  free-=child->free;
 }
 if(sb->RIGHT!=NIL_NODE)
 {
  child=(objstr*)((char*)sb+sb->RIGHT);
  free-=child->free;
 }
 return(free);
}


/* given there is already free bytes free in a node, set free for it */

static void tsetfree(objstr * sb,int free)
{
 objstr * child;

 if(sb->LEFT!=NIL_NODE)
 {
  child=(objstr*)((char*)sb+sb->LEFT);
  free+=child->free;
 }

 if(sb->RIGHT!=NIL_NODE)
 {
  child=(objstr*)((char*)sb+sb->RIGHT);
  free+=child->free;
 }

 sb->free=free;
}




/*
 * Balance the subtree rooted at sb that has become to heavy on side d
 * by single rotation of sb and sb_next.
 * Also adjusts sibling pointer of the parent of sb, or *rootp if sb is
 * the top of the entire tree.
 *
 *             sb              sb_next         Single rotation: Adding x or
 *            /  \            /       \        deleting under 3 caused
 *     sb_next    3          1         sb      rotation.  Same holds for mirror
 *     /       \             |        /  \     image.  Single_rotation returns
 *    1        2       ==>   x       2    3    top of new balanced tree.
 *    |                |                     |
 *    x                y                     y
 */



static objstr * single_rotation(objstr * sb,objstr * sb_next,DIRECT d,
                                                         domainstr * domain)
{
 int sbfree;

 sbfree=tfreenode(sb);
 sb_next->free=sb->free;

 *sibling(sb,d)=*sibling(sb_next,!d);
 if(*sibling(sb,d)!=NIL_NODE) *sibling(sb,d)+=((char*)sb_next-(char*)sb);

 *sibling(sb_next,!d)=(char*)sb-(char*)sb_next;

 sb->balance-=sb_next->balance;
 sb_next->balance=-sb->balance;

 *parentp(sb,domain)+=(char*)sb_next-(char*)sb;
 sb_next->parent=sb->parent;
 if(sb_next->parent!=NIL_NODE) sb_next->parent+=(char*)sb-(char*)sb_next;

 sb->parent=(char*)sb_next-(char*)sb;
 if(*sibling(sb,d)!=NIL_NODE) siblingp(sb,d)->parent=-(*sibling(sb,d));

 tsetfree(sb,sbfree);

 return(sb_next);
}




/*
 * Balance the subtree rooted at sb that has become to heavy on side d
 * by double rotation of sb and sb_next.
 * Also adjusts sibling pointer of the parent of sb, or *rootp if sb is
 * the top of the entire tree.
 *
 *             sb                  sb_next2       Double rotation: Adding x or
 *            /  \                /        \      x', or deleting under 4
 *     sb_next    \            sb_next      sb    caused rotation. Same holds
 *     /       \    \         /       \    /  \   for the mirror image.
 *    1   sb_next2   4  ==>  1         2  3    4  Double_rotation returns top
 *       /        \                    |  |       of new balanced tree.
 *      2          3                   x  x'
 *      |         |
 *      x         x'
 */

static objstr * double_rotation(objstr * sb,objstr * sb_next,DIRECT d,
                                                          domainstr * domain)
{
 int      sb_next2;
 objstr * sb_n2p;
 objstr * sb_np;
 objstr * sbp;
 int      sb_next_free;
 int      sb_free;

/* printf("double rot\n"); */

 sb_np=sb_next;
 sb_next2=*sibling(sb_np,!d);
 sb_n2p=(objstr*)((char*)sb_np+sb_next2);
 sbp=sb;

/* printf("sb=%x sb_n2p=%x sb_next=%x\n",(int)sb,(int)sb_n2p,(int)sb_next);
 printf("sbparent=%x\n",(int)((char*)sb+sb->parent));
*/

 sb_next_free=tfreenode(sb_next);
 sb_free=tfreenode(sb);
 sb_n2p->free=sb->free;


 *sibling(sb_np,!d) =*sibling(sb_n2p,d);
 if(*sibling(sb_np,!d)!=NIL_NODE)
                      *sibling(sb_np,!d)+=((char*)sb_n2p-(char*)sb_np);


 *sibling(sbp,d)    =*sibling(sb_n2p,!d);
 if(*sibling(sbp,d)!=NIL_NODE) 
                      *sibling(sbp,d)+=((char*)sb_n2p-(char*)sbp);



 *sibling(sb_n2p,d) =(char*)sb_next-(char*)sb_n2p;
 *sibling(sb_n2p,!d)=(char*)sb-(char*)sb_n2p;

 if(sb_n2p->balance==sb_np->balance) sb_np->balance=-sb_np->balance;
 else                                sb_np->balance=0;

 if(sb_n2p->balance==sbp->balance)   sbp->balance=-sbp->balance;
 else                                sbp->balance=0;

 sb_n2p->balance=0;

 *parentp(sb,domain)+=(char*)sb_n2p-(char*)sb;


 sb_n2p->parent=sbp->parent;
 if(sb_n2p->parent!=NIL_NODE) sb_n2p->parent-=(char*)sb_n2p-(char*)sb;

/* sbp->parent=sb_next2+(char*)sbp-(char*)sb_np; */
 sbp->parent=(char*)sb_n2p-(char*)sbp; 

 sb_np->parent=(char*)sb_n2p-(char*)sb_np;


 if(*sibling(sb_np,!d)!=NIL_NODE)
               siblingp(sb_np,!d)->parent=-(*sibling(sb_np,!d));

 if(*sibling(sbp,d) != NIL_NODE)
               siblingp(sbp,d)->parent=-(*sibling(sbp,d));

/*
 printf("\n\nsb=%x sbp->LEFT=%d sbp->RIGHT=%d\n",(int)sb_n2p,sb_n2p->LEFT,sb_n2p->RIGHT);
 printf("LEFT->parent=%d RIGHT->parent=%d\n",
                   ((objstr*)((char*)sb_n2p+sb_n2p->LEFT))->parent,
                   ((objstr*)((char*)sb_n2p+sb_n2p->RIGHT))->parent);

 */


 tsetfree(sb_next,sb_next_free);
 tsetfree(sb,sb_free);

 return(sb_n2p);
}



/*
 * Balance the subtree rooted at sb that has become to heavy on side d.
 * Also adjusts sibling pointer of the parent of sb, or *rootp if sb is
 * the top of the entire tree.
 */

static objstr * balance(objstr * sb,DIRECT d,domainstr * domain)
{
 objstr * sb_next;
  
 sb_next=siblingp(sb,d);

 if(sb_next->balance==-dir_bal(d))
 {
  return(double_rotation(sb,sb_next,d,domain));
 }
 else
 {
  return(single_rotation(sb,sb_next,d,domain));
 }
}






typedef enum
{
 preorder,
 postorder,
 endorder,
 leaf,
 root
} VISIT;




#ifdef NEVER

/*
 * Twalk walks the tree rooted by node, from top to bottom and left
 * to right, calling action with the NODE pointer, visit order, and
 * level in the tree (0 is root).  Leafs are visited only once and
 * action is then called with `leaf' as the second parameter.  For
 * nodes with children action is called three times with visit order
 * parameter `preorder' before, `postorder' in between, and `endorder'
 * after visiting the nodes children.
 */




static void twalk(domainstr * domain)
{
 VISIT    visit;
 int      level;
 objstr * node;
 objstr * parent;
 int      max;

 max=256;

 if(domain->root==NIL_NODE) return;
 node=(objstr*)(domain->domaindata+domain->root);

 /* run down tree from top to bottom, left to right */

/* printf("twalk"); */

 visit=preorder;
 level=0;
 while(level>=0 && max--)
 {
  if(visit==preorder && node->LEFT==NIL_NODE && node->RIGHT==NIL_NODE)
                                   visit=leaf; /* node turns out to be leaf */

  
  if(visit==leaf || visit==postorder) printf("node=%x",node->key);


  switch(visit)
  {
    case preorder:                  /* before visiting left child */
                  if(node->LEFT!=NIL_NODE)
                  {
                   node=(objstr*)((char*)node+node->LEFT);
                   level++;
                  }
                  else visit=postorder;
                  break;

   case postorder:         /* between visiting children */
                  if(node->RIGHT!=NIL_NODE)
                  {
                   node=(objstr*)((char*)node+node->RIGHT);
                   visit=preorder;
                   level++;
                  }
                  else visit=endorder;
                  break;

    case endorder:                      /* after visiting children */
        case leaf:                      /* node has no children */
                  if(node->parent!=NIL_NODE)
                  {
                   parent=(objstr*)((char*)node+node->parent);

                   if((node->parent!=-parent->LEFT) &&
                      (node->parent!=-parent->RIGHT))
                       printf("bad node key=%d",node->key);

                   if(direction(parent,node)==L) visit=postorder;
                   else                          visit=endorder;
                   node=parent;
                  }
                  level--;
                  break;
  }
 }
}


#endif







/*
 * Tfind searches node key in the tree rooted by *rootp, using compar for
 * comparing elements.  It returns the pointer to the NODE in which
 * the key pointer resides, or 0 if key is not present.
 */


static objstr * tfind(int key,domainstr * domain)
{
 objstr * nodep;
 int      node;
 int      cmp;

 node=domain->root;
 nodep=(objstr*)domain->domaindata;
 while(node!=NIL_NODE)
 {
  nodep=(objstr*)((char*)nodep+node);
  if((cmp=(key-nodep->key))==0) return(nodep);
  node=*sibling(nodep,cmp_dir(cmp));
 }

 return(NULL);
}





static objstr * tprev(objstr * obj)
{
 int delta;

 if(!obj) return(NULL);

//cprintf("tprev obj %x left %d right %d",(int)obj,obj->LEFT,obj->RIGHT);

 if(obj->LEFT!=NIL_NODE)
 {
  obj=(objstr*)((char*)obj+obj->LEFT);
  while(obj->RIGHT!=NIL_NODE)
  {
//cprintf("tprev1 obj %x left %d right %d",(int)obj,obj->LEFT,obj->RIGHT);
//	assert(obj->RIGHT!=0 && obj->LEFT!=0);
   obj=(objstr*)((char*)obj+obj->RIGHT);
  }
 }
 else
 {
  while(1)
  {
//cprintf("tprev2 obj %x left %d right %d parent %d",(int)obj,obj->LEFT,obj->RIGHT,obj->parent);

   if(obj->parent==NIL_NODE) return(NULL);

   delta=obj->parent;
   obj=(objstr*)((char*)obj+delta);

   if(delta==-obj->RIGHT) break;


//			assert(obj->RIGHT!=NIL_NODE && obj->LEFT!=NIL_NODE);
  }
 }
 return(obj);
}


static void tchangefree(objstr * prev,int free)
{
 while(1)
 {
/*  printf("prev=%x",prev); */
  prev->free+=free;
  if(prev->parent!=NIL_NODE) prev=(objstr*)((char*)prev+prev->parent);
  else                       break;
 }
}


/* how much free space in tree */

static int tfreetree(domainstr * domain)
{
 int      free;
 objstr * root;

 
	free=domain->bottom-sizeof(objfilehdr);
	if(root!=(objstr *)NIL_NODE)
	{
  root=(objstr*)(domain->domaindata+domain->root);
  free+=root->free;
	}

 return(free);
}


/* moves pointers to src to dest */

static void tmovenode(objstr * dest,objstr * src,domainstr * domain)
{
 objstr * child;
 objstr * sparent;
 int      sleft;
 int      sright;
 int    * p;

 sleft=src->LEFT;
 sright=src->RIGHT;

 dest->parent=src->parent;

 if(dest->parent!=NIL_NODE)
 {
  sparent=(objstr*)((char*)src+src->parent);
  p=sibling(sparent,direction(sparent,src));
  dest->parent+=(char*)src-(char*)dest;
  *p=-dest->parent;
 }
 else domain->root=(char*)dest-domain->domaindata;

 dest->LEFT=sleft;
 if(dest->LEFT!=NIL_NODE)
 {
  dest->LEFT+=(char*)src-(char*)dest;
  child=(objstr*)((char*)dest+dest->LEFT);
  child->parent=-dest->LEFT;
 }

 dest->RIGHT=sright;
 if(dest->RIGHT!=NIL_NODE)
 {
  dest->RIGHT+=(char*)src-(char*)dest;
  child=(objstr*)((char*)dest+dest->RIGHT);
  child->parent=-dest->RIGHT;
 }
}



/*
 * Tdelete removes key from the tree rooted by *rootp, using compar
 * for comparing elements.  It returns the pointer to the NODE that
 * was the parent of the NODE that contained the key pointer, 0 if
 * it did not exist.
 */


static os_error * tdelete(objstr * child,domainstr * domain)
{
 objstr * parent;
 objstr * dnode;
 objstr * dparent;
 objstr * rchild;
 DIRECT   d;
 int      cont_bal;
 int      free;

/* printf("tdelete(%d)",child->key); */

 /* the node was found; get its successor (if any) to replace it */

 dnode=child;

 if(child->RIGHT==NIL_NODE)
 {                                               /* no successor for key */
 
 /* printf("no right successor"); */

  if(child->LEFT!=NIL_NODE)
  {                                              /* set new parent */
  /* printf("left successor");  */

   child=(objstr*)((char*)dnode+child->LEFT);
   child->parent=dnode->parent;
   if(dnode->parent!=NIL_NODE) child->parent+=(char*)dnode-(char*)child;
  }

  if(dnode->parent==NIL_NODE)
  {
  /* printf("del root"); */
                                                 /*  free((char *) dnode); */
   if(dnode->LEFT==NIL_NODE) domain->root=NIL_NODE;
   else                      domain->root=(char*)child-domain->domaindata;

   return(NULL);                                 /* just deleted the root */
  }
  else dparent=(objstr*)((char*)dnode+dnode->parent);

 /* printf("dparent L=%d R=%d dnode P=%d",dparent->LEFT,dparent->RIGHT,
                                                                dnode->parent);
   */

  d=direction(dparent,dnode);                       /* for back up */

  if(dnode->LEFT!=NIL_NODE)
  {
   *sibling(dparent,d)=(char*)child-(char*)dparent; /* replace by left child */
  }
  else
   *sibling(dparent,d)=NIL_NODE;
 

 /* printf("dparent balance=%d left=%d right=%d d=%d\n",dparent->balance,
         dparent->LEFT,dparent->RIGHT,d); */

                                                 /* free ((char *) dnode); */
  parent=dparent;
 }
 else
 {
  child=(objstr*)((char*)child+child->RIGHT);
                                              /* key's successor exists */
  while(child->LEFT!=NIL_NODE)
  {
/*   printf("child-key=%d\n",child->key); */
   child=(objstr*)((char*)child+child->LEFT);
  }

/*  printf("*child->free=%d left=%d right=%d\n",child->free,child->LEFT,child->RIGHT); */
  free=tfreenode(child);
  tchangefree(child,-free);
/*  printf("*child=%d free=%d\n",child->free,free); */


  parent=(objstr*)((char*)child+child->parent);

 /* printf("child=%d parent=%d dnf=%d\n",child->key,parent->key,dnode->free);*/

/*  printf("parentfree=%d childfree=%d\n",tfreenode(parent),tfreenode(child));*/

  d=direction(parent,child);                    /* for back up */
  *sibling(parent,d)=child->RIGHT;
  if(child->RIGHT!=NIL_NODE)
  {
   *sibling(parent,d)+=(char*)child-(char*)parent;
   rchild=(objstr*)((char*)child+child->RIGHT);
   rchild->parent=(char*)parent-(char*)rchild;  /* -*sibling(parent,d) */

  /* printf("rchild=%d sibling=%d\n",rchild->parent,*sibling(parent,d)); */

  }

  /* now child, replaces dnode */
 /* printf("child replaces dnode"); */

  tmovenode(child,dnode,domain);
  child->balance=dnode->balance;
  child->free=dnode->free;
  tchangefree(child,free);

 /* printf("parentfree=%d\n",tfreenode(parent)); */

  if(parent==dnode) parent=child;

  /* dnode->key=child->key; */  /* successor replaces key */
  /*  free((char *) child); */
 }

        /*
         * Back up until tree is balanced.  This is achieved when either
         * the tree is balanced by rotation but not made shorter, or a
         * node's balance was 0 before deletion.

         */

 do
 {
 /* printf("parent=(%x),%d",(int)parent,parent->key); */

  if(parent->balance==dir_bal(!d))
  {
  /* printf("*c1*"); */
   cont_bal=((siblingp(parent,!d))->balance!=0);
   parent=balance(parent,!d,domain);
  }
  else
  {
  /* printf("parent->balance=%d +=%d\n",parent->balance,dir_bal(!d));*/
   cont_bal=(parent->balance!=0);
   parent->balance+=dir_bal(!d);
  }
  child=parent;

  if(parent->parent==NIL_NODE) break; /* we reached the root */
  parent=(objstr*)((char*)parent+parent->parent);
 /* printf("parent=(%x),%d cont=%d",(int)parent,parent->key,cont_bal); */

  d=direction(parent, child);
 } while(cont_bal); 

 return(NULL);
}




/*
 * Tsearch adds node key to tree rooted by *rootp, using compar for
 * comparing elements.  It returns the pointer to the NODE in which
 * the (possibly already existing) key pointer resides.
 */


static objstr * tsearch(objstr * nnode,domainstr * domain)
{
 objstr * parentp;
 int      child;
 DIRECT   d;
 int      cmp;
 objstr * childp;


 /* find place where key should go */

 parentp=NULL;
 cmp=0;              /* compiler */


 child=domain->root;
 childp=(objstr*)domain->domaindata;

 while(child!=NIL_NODE)
 {
  childp=(objstr*)((char*)childp+child);
  cmp=(nnode->key-childp->key);
  if(!cmp) return(childp);
  parentp=childp;
  child=*sibling(childp,cmp_dir(cmp));
 }

 /* create new node and set its parent's sibling pointer */

 nnode->balance=0;
 nnode->LEFT=nnode->RIGHT=NIL_NODE;

 if(parentp==NULL)
 {
  nnode->parent=NIL_NODE;
  domain->root=(char*)nnode-domain->domaindata;
  return(nnode);                   /* just created tree */
 }
 else nnode->parent=(char*)parentp-(char*)nnode;

 *sibling(parentp,cmp_dir(cmp))=(char*)nnode-(char*)parentp;
 childp=nnode;

 /*
  * Back up until tree is balanced.  This is achieved when either
  * the tree is balanced by rotation or a node's balance becomes 0.
  */

 while(1)
 {
  d=direction(parentp,childp);
  if(parentp->balance==dir_bal(d))
  {
   balance(parentp,d,domain);
   return(nnode);
  }
  else
  if((parentp->balance+=dir_bal(d))==0) return(nnode);

  if(parentp->parent==NIL_NODE) break;

  childp=parentp;
  parentp=(objstr*)((char*)parentp+parentp->parent);
 }

 return(nnode);
}



/****************************************************************************/

typedef struct twalkstr
{
 objstr * obj;
 VISIT    visit;
 DIRECT   direction;

} twalkstr;


static void objwalk(twalkstr * twalk)
{
 objstr * obj;
 objstr * parent;

 obj=twalk->obj;

 if(twalk->visit==preorder && obj->LEFT==NIL_NODE && obj->RIGHT==NIL_NODE)
                         twalk->visit=leaf; /* node turns out to be leaf */

  
/* if(twalk->visit==leaf || twalk->visit==postorder) dprintf(0,"node=%d",node->key); */



 if(twalk->direction==R) /* search goes left to right */
 {
  switch(twalk->visit)
  {
   case preorder:                  /* before visiting left child */
                 if(obj->LEFT!=NIL_NODE) obj=(objstr*)((char*)obj+obj->LEFT);
                 else                    twalk->visit=postorder;
                 break;

  case postorder:         /* between visiting children */
                 if(obj->RIGHT!=NIL_NODE)
                 {
                  obj=(objstr*)((char*)obj+obj->RIGHT);
                  twalk->visit=preorder;
                 }
                 else twalk->visit=endorder;
                 break;

   case endorder:                      /* after visiting children */
       case leaf:                      /* node has no children */
                 if(obj->parent!=NIL_NODE)
                 {
                  parent=(objstr*)((char*)obj+obj->parent);

                  if(direction(parent,obj)==L) twalk->visit=postorder;
                  else                         twalk->visit=endorder;
                  obj=parent;
                 }
                 else twalk->visit=root;
                 break;
  }
 }
 else   /* L search goes right to left */
 {
  switch(twalk->visit)
  {
   case preorder:                  /* before visiting left child */
                 if(obj->RIGHT!=NIL_NODE) obj=(objstr*)((char*)obj+obj->RIGHT);
                 else                    twalk->visit=postorder;
                 break;

  case postorder:         /* between visiting children */
                 if(obj->LEFT!=NIL_NODE)
                 {
                  obj=(objstr*)((char*)obj+obj->LEFT);
                  twalk->visit=preorder;
                 }
                 else twalk->visit=endorder;
                 break;

   case endorder:                      /* after visiting children */
       case leaf:                      /* node has no children */
                 if(obj->parent!=NIL_NODE)
                 {
                  parent=(objstr*)((char*)obj+obj->parent);

                  if(direction(parent,obj)==R) twalk->visit=postorder;
                  else                         twalk->visit=endorder;
                  obj=parent;
                 }
                 else twalk->visit=root;
                 break;
  }
 }


 if(twalk->visit==preorder && obj->LEFT==NIL_NODE && obj->RIGHT==NIL_NODE)
                         twalk->visit=leaf; /* node turns out to be leaf */




 twalk->obj=obj;
}







/* try to find free bytes in the tree */

static void findfree(objstr * obj,objstr ** objlop,objstr ** objhip,int free,
                                                                 domainstr * d)
{
 twalkstr lo;
 twalkstr hi;
 int      freelo;
 int      freehi;
 int      delta;
 objstr * objlo;
 objstr * objhi;


 objlo=lo.obj=obj;
 lo.direction=L;
 lo.visit=postorder;

 objhi=hi.obj=obj;
 hi.direction=R;
 hi.visit=postorder;

 freelo=0;
 freehi=0;

 /* objlo scans up the tree   */
 /* objhi scans down the tree */

 while((freelo+freehi)<free)
 {
/*printf("lo=%x hi=%x fl=%d fh=%d\n",objlo->key,objhi->key,freelo,freehi);*/

  if(((obj-lo.obj)>(hi.obj-obj) || lo.visit==root) && hi.visit!=root)
  {
   objwalk(&hi);
   if(hi.obj->free==0) hi.visit=endorder;
   else
   if(hi.visit==leaf || hi.visit==postorder)
   {
    delta=tfreenode(hi.obj);
    if(delta)
    {
     freehi+=delta;
     objhi=hi.obj;
    }
   }
  }
  else
  if(lo.visit!=root)
  {
   objwalk(&lo);
   if(lo.obj->free==0) lo.visit=endorder;
   else
   if(lo.visit==leaf || lo.visit==postorder)
   {
    delta=tfreenode(lo.obj);
    if(delta)
    {
     freelo+=delta;
     objlo=lo.obj;
    }
   }
  }
  else break;
 }

 *objlop=objlo;
 *objhip=objhi;

 USE(d);
}




static void lockblock(objstr * objlo,objstr * objhi,int shift,domainstr * d)
{
 int       lockn;
 lockstr * lock;
 objstr  * obj;
 int       i;

 lockn=d->lockn;
 lock=d->lock;

 for(i=0;i<lockn;i++)
 {
  obj=(objstr*)lock->obj;
  if(obj>objlo && obj<=objhi)
  {
   obj=(objstr*)((char*)obj+shift);
   lock->obj=obj;
   *(lock->ptr)=&obj->key;
  }
  lock++;
 }
}


static void lockinsert(objstr * objpt,int shift,domainstr * d)
{
 int       lockn;
 lockstr * lock;
 objstr  * obj;
 int       i;

 lockn=d->lockn;
 lock=d->lock;

 for(i=0;i<lockn;i++)
 {
  obj=(objstr*)lock->obj;
  if(obj>objpt)
  {
   obj=(objstr*)((char*)obj+shift);
   lock->obj=obj;
   *(lock->ptr)=&obj->key;
  }
  lock++;
 }
}



static os_error * lockdomain(int handle,int by)
{
 domainstr * d;
 int         lockn;
 lockstr   * lock;
 objstr    * obj;
 int         i;

//cprintf("lock domain handle=%x by=%d\n",handle,by);

 d=(domainstr *)handle;

 lockn=d->lockn;
 lock=d->lock;

/* printf("lockn=%d\n",lockn);  */

 for(i=0;i<lockn;i++)
 {
  obj=(objstr*)lock->obj;
  obj=(objstr*)(((char*)obj)+by);
  lock->obj=obj;
  *(lock->ptr)=&obj->key;

//cprintf("key=%d obj->key=%d",lock->key,obj->key);

  lock++;
 }
/* printf("lock domain x\n"); */

 return(NULL);
}



/* move objects from objlo to objhi+length(objhi) by shift */
/* objects below objlo and including objlo remain fixed    */
/* objects above objhi remain fixed                        */
/* anything >objlo and <=objhi gets moved                  */

static void moveblock(objstr * objlo,objstr * objhi,int shift,domainstr * d)
{
 objstr * child;
 objstr * sblo;
 objstr * sbhi;
 objstr * lleft;
 objstr * lright;
 objstr * hleft;
 objstr * hright;
 objstr * root;
 int      misslo;
 int      misshi;


/*
 printf("move block lo=%x hi=%x shift=%d\n",objlo->key,objhi->key,shift);
 */

 sblo=sbhi=root=(objstr*)(d->domaindata+d->root);
 misslo=misshi=0;
 lleft=lright=NULL;  /* make compiler happy */

 if(root>objlo && root<=objhi)
 {
  d->root+=shift;
 }

 while(1)
 {
  if(!misslo)
  {
   if(sblo->RIGHT!=NIL_NODE)
   {
    lright=child=(objstr*)((char*)sblo+sblo->RIGHT);
    if(child>objlo && child<=objhi && sblo<=objlo)
    {
     child->parent-=shift;
     sblo->RIGHT+=shift;
    }
    else
    if(sblo>objlo && sblo<=objhi && child>objhi)
    {
     child->parent+=shift;
     sblo->RIGHT-=shift;
    }
   }
   else lright=NULL;

   if(sblo->LEFT!=NIL_NODE)
   {
    lleft=child=(objstr*)((char*)sblo+sblo->LEFT);
    if(child<=objlo && sblo>objlo && sblo<=objhi)
    {
     child->parent+=shift;
     sblo->LEFT-=shift;
    }
    else
    if(child>objlo && child<=objhi && sblo>objhi)
    {
     child->parent-=shift;
     sblo->LEFT+=shift;
    }
   }
   else lleft=NULL;
  }

  if(sblo!=sbhi && !misshi)
  {
   if(sbhi->RIGHT!=NIL_NODE)
   {
    hright=child=(objstr*)((char*)sbhi+sbhi->RIGHT);
    if(child>objlo && child<=objhi && sbhi<=objlo)
    {
     child->parent-=shift;
     sbhi->RIGHT+=shift;
    }
    else
    if(sbhi>objlo && sbhi<=objhi && child>objhi)
    {
     child->parent+=shift;
     sbhi->RIGHT-=shift;
    }
   }
   else hright=NULL;

   if(sbhi->LEFT!=NIL_NODE)
   {
    hleft=child=(objstr*)((char*)sbhi+sbhi->LEFT);
    if(child<=objlo && sblo>objlo && sbhi<=objhi)
    {
     child->parent+=shift;
     sbhi->LEFT-=shift;
    }
    else
    if(child>objlo && child<=objhi && sbhi>objhi)
    {
     child->parent-=shift;
     sbhi->LEFT+=shift;
    }
   }
   else hleft=NULL;
  }
  else
  {
   hleft=lleft;
   hright=lright;
  }

  if(sblo==objlo) misslo=1;
  if(sbhi==objhi) misshi=1;
  if(misshi && misslo) break;

  if(!misslo)
  {
   if(objlo<sblo) sblo=lleft;
   else           sblo=lright;
  }

  if(!misshi)
  {
   if(objhi>sbhi) sbhi=hright;
   else           sbhi=hleft;
  }

/*  printf("sblo=%x sbhi=%x oblo=%x obhi=%x\n",sblo,sbhi,objlo,objhi); */

  if(!sblo || !sbhi) break; /* oops something is radically wrong */
 }


 objlo->length+=shift;
 objhi->length-=shift;

 if(shift<0)
 {
  memmove((char*)objlo+objlo->length,(char*)objlo+objlo->length-shift,
        ((char*)objhi+objhi->length+shift)-
        ((char*)objlo+objlo->length-shift));
 }
 else
 {
  memmove((char*)objlo+objlo->length,(char*)objlo+objlo->length-shift,
        ((char*)objhi+objhi->length)-
        ((char*)objlo+objlo->length-shift));
 }

 lockblock(objlo,objhi,shift,d);
}





/* insert delta bytes at the end of obj              */
/* assume domain is big enough and increase its size */

static void insertblock(objstr * obj,int delta,domainstr * d)
{
 objstr * child;
 objstr * next;
 objstr * sb;
 int      offset;


/* printf("insert block %d\n",delta); */

 next=(objstr *)((char*)obj+obj->length);

 sb=obj;

 while(1)
 {
  if(sb->RIGHT!=NIL_NODE)
  {
   child=(objstr*)((char*)sb+sb->RIGHT);
   if(child>=next && sb<next)
   {
    child->parent-=delta;
    sb->RIGHT+=delta;
   }
  }

  if(sb->LEFT!=NIL_NODE)
  {
   child=(objstr*)((char*)sb+sb->LEFT);
   if(child<next && sb>=next)
   {
    child->parent+=delta;
    sb->LEFT-=delta;
   }
  }

  if(sb->parent==NIL_NODE) break;
  sb=(objstr*)((char*)sb+sb->parent);
 }



 /* printf("domain=%x lockn=%d data=%x %x\n",(int)d,d->lockn,(int)d->domaindata,d->domaindata[-16]); */



 /* do it */

 offset=(char*)next-d->domaindata;
 if(d->root>=offset) d->root+=delta;
 memmove((char*)next+delta,(char*)next,d->size-offset);

/* printf("memmove %x %x %d\n",(int)((char*)next+delta),
                             (int)((char*)next),
                           d->size-offset);     */

/*  printf("domain=%x lockn=%d data=%x %x\n",(int)d,d->lockn,(int)d->domaindata,d->domaindata[-16]);  */


 d->size+=delta;
 obj->length+=delta;
 lockinsert(obj,delta,d);
}



/* compact starting with objlo and going upwards till you hit obj */

static int compactdown(objstr * objlo,objstr * obj,domainstr * d)
{
 int      freelo=0;
 int      freesum;
 int      free;
 objstr * sb;

/* printf("compact down lo=%x ob=%x\n",(int)objlo,(int)obj); */

 freesum=0;

 while(1)
 {
  freelo=tfreenode(objlo);
  if(freelo || objlo==obj) break;
  objlo=(objstr*)((char*)objlo+objlo->length);
 }

 sb=objlo;

 while(1)
 {

  if(sb==obj) break;
  sb=(objstr*)((char*)sb+sb->length);

  free=tfreenode(sb);
  if(free || sb==obj)
  {
   moveblock(objlo,sb,-freelo,d);

   if(sb==obj)
   {
    freesum=-freelo;
    obj=(objstr*)((char*)sb-freelo);
   }
   sb=(objstr*)((char*)sb-freelo);

   tchangefree(objlo,-freelo);
   tchangefree(sb,freelo);
   objlo=sb;
   freelo+=free;
  }
 }

 return(freesum);
}


/* compact from objhi and go down till you hit obj */
/* 'new' bytes have been added to the domain size  */

static void compactup(objstr * objhi,objstr * obj,domainstr * d)
{
 int      freehi=0;
 int      free;
 objstr * sb;

// cprintf("compact up objhi %x obj %x",(int)objhi,obj);



 while(1)
 {
  freehi=tfreenode(objhi);
  if(freehi || objhi==obj) break;
  objhi=tprev(objhi);
 }

 sb=objhi;




 while(1)
 {
//		cprintf("sb %x obj %x",sb,obj);

  if(sb==obj) break;
  sb=tprev(sb);
  free=tfreenode(sb);
  if(free || sb==obj)
  {
   moveblock(sb,objhi,freehi,d);
   objhi=(objstr*)((char*)objhi+freehi);
   tchangefree(objhi,-freehi);
   tchangefree(sb,freehi);
   objhi=sb;
   freehi+=free;
  }
 }
}



static void compactdomain(domainstr * d)
{
 objstr   * objlo;
 objstr   * objhi;

 objlo=(objstr*)(d->domaindata+d->bottom);
 objhi=tfind(d->maxobject,d);

 compactdown(objlo,objhi,d);

 objhi=tfind(d->maxobject,d);
 objhi->length-=objhi->free;
 d->size-=objhi->free;
 tchangefree(objhi,-objhi->free);
}


os_error * ob_compactdomain(domainstr * d)
{
 compactdomain(d);
 return(dextend(d,d->size));
}



static os_error * ob_compactallx(domainstr * x)
{
 os_error  * err;
 domainstr * d;

 err=NULL;

 d=firstdomain;
 while(d)
 {
  if(d!=x)
  {
   err=ob_compactdomain(d);
   if(err) break;
  }
  d=d->next;
 }

 return(err);
}



os_error * ob_compactall(void)
{
 return(ob_compactallx(NULL));
}




os_error * ob_checkfree(int need,domainstr * d)
{
 os_error * err;
 int        free;         /* free memory in the tree */
 int        xnew;          /* free memory added to domain      */
 objstr   * objhi;


 err=NULL;

 free=tfreetree(d);
 if(free<need)
 {
  xnew=need-free;

  err=dchunk(d,xnew,NULL);
  if(!err)
  {
   d->size+=xnew;
   objhi=tfind(d->maxobject,d);
   tchangefree(objhi,xnew);
   objhi->length+=xnew;
  }
 }

 return(err);
}





static os_error * expandobj(objstr ** objp,int extra,domainstr * d)
{
 os_error * err;
 objstr   * obj;
 int        key;
 int        free;         /* free memory in the tree */
 /* int        tfree; */  /* free memory coming from the tree */
 int        xnew;          /* free memory added to domain      */
 objstr   * objlo=NULL;
 objstr   * objhi=NULL;
 int        cup;
 int        ofree;        /* how much already free in this object */

 obj=*objp;
 key=obj->key;

 err=NULL;

 ofree=tfreenode(obj);
 free=tfreetree(d);
 free=free-ofree;        /* how much free outside this object    */


 if(free)
 {
  if(extra>free) /* have to expand domain */
  {
   xnew=extra-free;
   err=dchunk(d,xnew,&obj);
   if(!err)
   {
    d->size+=xnew;
    objlo=(objstr*)(d->domaindata+d->bottom);
    objhi=tfind(d->maxobject,d);
    tchangefree(objhi,xnew);
    objhi->length+=xnew;
   }
  }
  else
  {
   findfree(obj,&objlo,&objhi,extra,d); /* find extra in tree */
  }

  if(!err)
  {
   cup=objhi>obj;
   if(objlo<obj)
   {
    obj=(objstr*)((char*)obj+compactdown(objlo,obj,d));
   }
   if(cup)
   {
    compactup(objhi,obj,d);
   }
 /*  tchangefree(obj,-extra); */
  }
 }
 else
 {
  err=dchunk(d,extra,NULL);
  if(!err)
  {
   obj=tfind(key,d);
   insertblock(obj,extra,d);
  }
 }

 *objp=obj;

 return(err);
}



/*****************************************************************************/



void * ob_find(int number,domainstr * domain)
{
 objstr * obj;
 /* dprintf(0,"ob find");  */

 obj=tfind(number,domain);

#ifdef _DEBUG
//	assert(obj!=NULL);
#endif

 if(obj)
 {

/* dprintf(0,"ob findx");  */

  return(&obj->key);
 }
 else return(NULL);
}


os_error * ob_change(int n,domainstr * d,int newsize,objptr pointer)
{
 os_error * err;
 objstr   * obj;
 int        free;
 int        oldsize;
 int        extra;

 err=NULL;

 newsize+=OVERHEAD;
 newsize=(newsize+0x3)&(~0x3); /* objects are word aligned */

 obj=tfind(n,d);
 if(obj->type==MIRROR)
 {




 }
 else
 {
  if(obj->length>=newsize)
  {
   free=tfreenode(obj);
   oldsize=obj->length-free;
   if(oldsize!=newsize) tchangefree(obj,oldsize-newsize);
  }
  else  /* need to expand object */
  {
   extra=newsize-obj->length; /* we need this many bytes extra */

   err=expandobj(&obj,extra,d);

   free=tfreenode(obj);
   oldsize=obj->length-free;
   tchangefree(obj,oldsize-newsize);
  }
 }

 *pointer=&obj->key;
 return(err);
}



os_error * ob_mod(int n,domainstr * d,objptr pointer)
{
 os_error * err;
 objstr   * obj;

 err=NULL;

 obj=tfind(n,d);
 if(obj->type==MIRROR)
 {




 }

 if(obj) *pointer=&obj->key;
	else				*pointer=NULL;

 return(err);
}



os_error * ob_setbits(int n,domainstr * d,int bits,objptr pointer)
{
 os_error * err;
 objstr   * obj;

 err=NULL;

 obj=tfind(n,d);

 obj->copyclass=bits;

 *pointer=&obj->key;
 return(err);
}




static os_error * newobjectspace(int size,domainstr * d)
{
 os_error * err;
 objstr   * objlo;
 objstr   * objhi;
 int        free;
 int        extra;


 err=NULL;
 free=tfreetree(d);

 if(free)
 {
  objhi=tfind(d->maxobject,d);
  free=tfreenode(objhi);

  if(free<size)
  {
   findfree(objhi,&objlo,&objhi,size,d);
   compactdown(objlo,objhi,d);
   objhi=tfind(d->maxobject,d);
   free=tfreenode(objhi);
  }

  objhi->length-=free;
  d->size-=free;
  tchangefree(objhi,-free);
  extra=size-free;

  if(extra>0)
  {
   err=dchunk(d,size,NULL); /* was size */
                       /* need to make current d->size greater by size    */
                       /* d->size is now free smaller than it was         */
  }
 }
 else
 {
  err=dchunk(d,size,NULL);
 }

 return(err);
}




os_error * ob_create(int size,domainstr * d,int type,objptr pointer)
{
 os_error * err;
 objstr   * nnode;

 size+=OVERHEAD;
 size=(size+0x3)&(~0x3);

 err=newobjectspace(size,d);
 if(!err)
 {
  nnode=(objstr*)(d->domaindata+d->size);
  memset(nnode,0,size);
  d->size+=size;
  nnode->key=++d->maxobject;
  nnode->magic=DMMAGIC;
  nnode->length=size;
  nnode->type=type;
  nnode=tsearch(nnode,d);
  *pointer=&nnode->key;
 }
 return(err);
}




os_error * ob_delete(int n,domainstr * d)
{
 os_error * err;
 objstr   * obj;
 objstr   * prev;
 int        length;
 int        free;

 err=NULL;


 if(n)
 {
  obj=tfind(n,d);
  if(obj)
  {
 /*  dprintf(2,"count=%d",obj->count); */

   if(obj->copyclass!=OBJ_CCINSTANCE || obj->count--<=0)
   {
    length=obj->length;

    free=tfreenode(obj);

    prev=tprev(obj);

    tchangefree(obj,-free);

    err=tdelete(obj,d);

    if(n==d->maxobject)
    {
     d->size-=length;
     if(prev) d->maxobject=prev->key;
     else
     {
      /* there's nothing left */
      if(d->bottom)
      {
       d->size+=d->bottom-sizeof(objfilehdr);
       d->bottom=sizeof(objfilehdr);
      }
     }
    }
    else
    {
     if(prev)
     {
      prev->length+=length;
     /* prev->count+=length; */
      tchangefree(prev,length);
     }
     else     d->bottom+=length;
    }
   }
  }
 }
 return(err);
}



os_error * ob_lock(void ** ptr,int n,domainstr * d)
{
 os_error * err;
 int        lockn;
 lockstr  * lock;
 objstr   * obj;

 err=NULL;

 lockn=d->lockn;
 lock=&d->lock[lockn];

 lock->key=n;
 lock->ptr=ptr;
 obj=tfind(n,d);

#ifdef _DEBUG

 assert(obj!=NULL);

#endif

 if(obj)
 {
  lock->obj=obj;
  *ptr=&obj->key;

  d->lockn++;
 }
 else
 {
  *ptr=NULL;
  err=geterror(EOBJECT00);
 }

 return(err);
}



os_error * ob_unlock(void ** ptr,domainstr * d)
{
 int       lockn;
 lockstr * lock;
 int       i;

 lockn=d->lockn;
 lock=d->lock;

 for(i=0;i<lockn;i++)
 {
  if(lock[i].ptr==ptr)
  {
   lock[i]=lock[lockn-1];
   d->lockn--;
   *ptr=NULL;
  }
 }
 return(NULL);
}




os_error * ob_copy(void ** dptr,domainstr * dd,int sn,domainstr * sd)
{
 os_error * err;
 void     * sptr;
 objstr   * soptr;
 objstr   * doptr;
 int        size;
 int        type;

 err=ob_lock(&sptr,sn,sd);
 if(!err)
 {
  soptr=(objstr*)((char*)sptr-OVERHEAD);

   cprintf("copyclass %x CCINS %x dd==sd",soptr->copyclass,OBJ_CCINSTANCE,dd==sd);

  if(soptr->copyclass==OBJ_CCINSTANCE && dd==sd)
  {
   soptr->count++;
   *dptr=sptr;
  }
  else
  {
   size=soptr->length-OVERHEAD-tfreenode(soptr);
   type=soptr->type;

   err=ob_create(size,dd,type,dptr);
   if(!err)
   {
    doptr=(objstr*)((char*)(*dptr)-OVERHEAD);

    size-=2*sizeof(int);   /* key + type !!! */
 
    soptr=(objstr*)((char*)sptr-OVERHEAD);
    XMEMCPYX(doptr->data,soptr->data,size);
    doptr->copyclass=soptr->copyclass;
   }
  }
  ob_unlock(&sptr,sd);
 }

 return(err);
}



int ob_size(int n,domainstr * d,objptr pointer)
{
 objstr   * obj;

 obj=tfind(n,d);
 if(obj)
 {
  *pointer=&obj->key;
  return(obj->length-OVERHEAD-tfreenode(obj));
 }
 return(-1);
}



/* allocates enough static memory to hold a copy of the object */
/* then copies the object into it                              */

os_error * ob_static(void ** ptr,int n,domainstr * d)
{
 os_error * err;
 void     * sptr;
 objstr   * soptr;
 int        size;

 err=ob_lock(&sptr,n,d);
 if(!err)
 {
  soptr=(objstr*)((char*)sptr-OVERHEAD);
  size=soptr->length-OVERHEAD-tfreenode(soptr);

  err=ralloc(ptr,size);
  XMEMCPYX(*ptr,sptr,size);

  ob_unlock(&sptr,d);
 }
 return(err);
}



os_error * ob_free(void ** ptr,domainstr * d)
{
 os_error * err;

 err=sfree(ptr);

 return(err);

 USE(d);
}



/*****************************************************************************/
/* whole domain operations */


static os_error * ob_createdomainsize(domainstr ** dp,int size)
{
 os_error  * err;
 domainstr * d;

 err=salloc((flex_ptr)dp,sizeof(domainstr));

 if(!err)
 {
  d=*dp;

  d->root=NIL_NODE;
  d->maxobject=FIRSTOBJECT;
  d->bottom=sizeof(objfilehdr);
  d->size=sizeof(objfilehdr);
  d->lockn=0;

  if(firstdomain) firstdomain->prev=d;
  d->next=firstdomain;
  firstdomain=d;

  err=dalloc(d,MAX(size,sizeof(objfilehdr)));
  if(!err) memset(d->domaindata,0,sizeof(objfilehdr));

  if(!err) err=flex_setlock((flex_ptr)&d->domaindata,lockdomain,(int)d);

  if(err) ob_deletedomain(dp);
 }

 return(err);
}



os_error * ob_createdomain(domainstr ** d)
{
 return(ob_createdomainsize(d,DMCHUNK));
}


os_error * ob_deletedomain(domainstr ** dp)
{
 os_error  * err;
 domainstr * prev;
 domainstr * next;
 domainstr * d;

 err=NULL;


 if((d=(*dp))!=NULL)
 {
  dfree(d);

  prev=d->prev;
  next=d->next;

  if(next) next->prev=prev;
  if(prev) prev->next=next;
  if(firstdomain==d) firstdomain=next;

  err=sfree((flex_ptr)dp);
 }

 return(err);
}



#ifndef MEMTEST




static os_error * ob_checkhdr(objfilehdr * hdr)
{
 char         mark[8];

	if(strcmp(hdr->progname,wimpt_programname())) return(geterror(EFILETYPE));

 if(hdr->compatability>maxcompatability) return(geterror(ECOMPAT));
 else
 {
  sprintf(mark,"%d.%02d",minversion/100,minversion % 100);

  if(strcmp(hdr->progmark,mark)<0) return(geterror(EVERSION));

		if(hdr->compatability<MINCOMPATABILITY) return(geterror(EVERSION));
 }

 return(NULL);
}



os_error * ob_loadfile(char * name,domainstr ** d,int * object,int flags,int * xcompatability)
{
 os_error   * err;
 objfilehdr * hdr;
 int          n;
 int          chunk;
 int          read;
 int          eof;
 int          first;
 int          crc;


 err=NULL;
 first=1;
 n=0;

 err=bf_loadstart(name);
 do
 {
  err=bf_loadsize(&chunk);
  if(err) break;

  if(first)
  {
   err=ob_createdomainsize(d,n+chunk);
   if(err) break;
   first=0;
  }
  else
  {
   err=dextend((*d),n+chunk);
   if(err) break;
  }

  err=bf_loaddata(name,(*d)->domaindata+n,chunk,&read,&eof);
  if(err) break;
  n+=read;
 } while(!eof);

 if(!err)
 {
//  cprintf("n=%d",n);

  if(n<sizeof(objfilehdr)) err=geterror(EFILETYPE);
  else
  {
   hdr=(objfilehdr *)(*d)->domaindata;
   err=ob_checkhdr(hdr);
   if(!err)
   {
    (*d)->root=hdr->root;
    (*d)->maxobject=hdr->maxobject;
    (*d)->size=hdr->size;            /* can check this */

/* dprintf(0,"hdrsize=%d n=%d",hdr->size,n); */

  /* (*d)->size=n; */

    *object=hdr->object;
				*xcompatability=hdr->compatability;

    if(flags)
    {
     if(n!=hdr->size) err=geterror(ETRUNCF);
     else
     {
      crc=hdr->crc;
      if(crc && (strcmp(hdr->progmark,"2.68")>0))
      {
       hdr->crc=0;
       crcblock32(&hdr->crc,(*d)->domaindata,(*d)->size);
       if(hdr->crc!=crc) err=geterror(ECRCF);
      }
     }
    }
   }
  }
 }

 if(err)
 {
  if(!first) ob_deletedomain(d);
 }

 return(err);
}






os_error * ob_readfile(buffer * bf,objfilehdr * hdr,domainstr ** d,
                                                    int * object)
{
 os_error   * err;

 err=ob_checkhdr(hdr);
 if(!err)
 {
  err=ob_createdomainsize(d,hdr->size);
  if(!err)
  {
   bf_seek(bf,0);
   err=bf_read(bf,(*d)->domaindata,hdr->size);
   if(err) ob_deletedomain(d);
   else
   {
    (*d)->root=hdr->root;
    (*d)->maxobject=hdr->maxobject;
    (*d)->size=hdr->size;            /* can check this */

    *object=hdr->object;
   }
  }
 }

 return(err);
}






os_error * ob_savefile(char * name,domainstr * d,int object,int ftype,int flags,int xcompatability)
{
 os_error   * err;
 objfilehdr * hdr;
 objstr     * obj;

 compactdomain(d);

 hdr=(objfilehdr *)d->domaindata;

 xzstrncpy(hdr->progname,wimpt_programname(),sizeof(hdr->progname));
 sprintf(hdr->progmark,"%d.%02d",version/100,version % 100);
 sprintf(hdr->progdate,datestring);

 obj=(objstr*)ob_find(object,d);

 hdr->version=version;
 hdr->compatability=xcompatability;
 hdr->object=object;
 hdr->type=obj->type;
 hdr->os=1;
 hdr->crc=0;
 hdr->root=d->root;
 hdr->maxobject=d->maxobject;
 hdr->size=d->size;

 if(flags)
 {
  crcblock32(&hdr->crc,d->domaindata,d->size);
 }

// cprintf("save block %s",name);

 err=bf_saveblock(name,ftype,hdr,d->size);

 return(err);
}


#endif


/*****************************************************************************/


os_error * ob_init(void)
{
 ofb=&ofbb;
 return(flex_createarea(&ofb,FLEX_NORESIZE,FLEX_MAX,wimpt_programname()));
}




os_error * ob_test(domainstr * d)
{
 objstr * maxobj;
 objstr * obj;
 objstr * lastobj;
 objstr * firstobj;
 objstr * child;
 objstr * root;
 objstr * parent;
 int      freenode;
 int      freetotal;
 int      balance;


 firstobj=lastobj=obj=(objstr*)(d->domaindata+d->bottom);
 maxobj=(objstr*)(d->domaindata+d->size);
 freetotal=0;

//	cprintf("test obj %x maxobj %x",(int)obj,(int)maxobj);

 if(d->size>d->bottom)
 {
  while(1)
  {
   if(obj->magic!=DMMAGIC)
   {
    return(generror(0,"Bad magic number at obj=%x (%d)",
                                (int)obj,((int)obj)-((int)firstobj)));
   }


			if(obj->parent!=NIL_NODE)
			{
				parent=(objstr*)((char*)obj+obj->parent);

    if(parent<firstobj || parent>maxobj)
    {
     return(generror(0,"Bad parent at obj=%x child=%x %d\n",
                      (int)obj,(int)child,((int)child)-((int)firstobj)));
    }


				if(parent->LEFT==NIL_NODE && parent->RIGHT==NIL_NODE)
				{
					return(generror(0,"parent does not point to child obj=%x",(int)obj));

				}

			}

			if(obj->LEFT==0 || obj->RIGHT==0)
			{
				return(generror(0,"object with null left or right child obj=%x",(int)obj));

			}



   freenode=obj->free;
   balance=obj->balance;

   if(obj->LEFT!=NIL_NODE)
   {
    child=(objstr*)((char*)obj+obj->LEFT);

    if(child<firstobj || child>maxobj)
    {
     return(generror(0,"Bad left child at obj=%x child=%x %d\n",
                      (int)obj,(int)child,((int)child)-((int)firstobj)));
    }
    else
    {
     if(child->parent!=-obj->LEFT)
     {
      return(generror(0,"Bad left offset at obj=%x child=%x %d %d\n",
                         (int)obj,(int)child,child->parent,obj->LEFT));

     }
     if(child->magic!=DMMAGIC)
     {
      return(generror(0,"Bad magic child at obj=%x child=%x %d\n",
                         (int)obj,(int)child,((int)child)-((int)firstobj)));
     }
     freenode-=child->free;
    }
   }                                                                           

   if(obj->RIGHT!=NIL_NODE)
   {
    child=(objstr*)((char*)obj+obj->RIGHT);
 
    if(child<firstobj || child>maxobj)
    {
     return(generror(0,"Bad right child at obj=%x child=%x %d\n",
            (int)obj,(int)child,((int)child)-((int)firstobj)));
    }
    else
    {
     if(child->parent!=-obj->RIGHT)
     {
      return(generror(0,"Bad right offset at obj=%x child=%x %d %d\n",
             (int)obj,(int)child,child->parent,obj->RIGHT));
     }
  
     if(child->magic!=DMMAGIC)
     {
      return(generror(0,"Bad magic child at obj=%x child=%x %d\n",
                (int)obj,(int)child,((int)child)-((int)firstobj)));
     }

     freenode-=child->free;
    }
   }

   if(freenode>obj->length)
   {
    return(generror(0,"Object has more free than length at obj=%x key=%d\n",
           (int)obj,obj->key));
   }


   if(
      (obj->LEFT  && !obj->RIGHT && obj->balance!=1) ||
      (obj->RIGHT && !obj->LEFT && obj->balance!=-1))
   {
    return(generror(0,"Bad balance at obj=%x key=%d\n",(int)obj,obj->key));
   }

   freetotal+=freenode;

   lastobj=obj;

   obj=(objstr*)((char*)obj+obj->length);
 
   if(obj>=maxobj) break;
  }
 }

 if(obj>maxobj)
 {
  return(generror(0,"Object pointed past end of domain at obj=%x\n",(int)obj));
 }

 if(lastobj->key!=d->maxobject)
 {
  return(generror(0,"Max objects wrong %d %d\n",lastobj->key,d->maxobject));
 }

 root=(objstr*)(d->domaindata+d->root);

 if(root->free!=freetotal)
 {
  return(generror(0,"Disagreement of free space %d %d",root->free,freetotal));
 }

 return(NULL);
}





os_error * ob_enum(domainstr * d,obenumfn fn)
{
 os_error * err;
 objstr   * maxobj;
 objstr   * obj;


 err=NULL;


 obj=(objstr*)(d->domaindata+d->bottom);
 maxobj=(objstr*)(d->domaindata+d->size);

 while(1)
 {
  err=fn(&obj->key);
  if(err) break;

  obj=(objstr*)((char*)obj+obj->length);
 
  if(obj>=maxobj) break;
 }

 return(err);
}



#ifndef PROD

void ob_testtree(domainstr * d)
{
 objstr * maxobj;
 objstr * obj;
 objstr * lastobj;
 objstr * firstobj;
 objstr * child;
 objstr * root;
 int      freenode;
 int      freetotal;
 int      balance;
 int      leftb;
 int      rightb;

 dprintf(0,"** Test Tree **\n");

 firstobj=lastobj=obj=(objstr*)(d->domaindata+d->bottom);
 maxobj=(objstr*)(d->domaindata+d->size);
 freetotal=0;

 dprintf(0,"first=%x max=%x",firstobj,maxobj);

 if(d->size>d->bottom)
 {
 while(1)
 {
  dprintf(0,"obj @ %d  obj->length=%d key=%d type=%d",
               ((int)obj)-((int)firstobj),obj->length,obj->key,obj->type); 


  if(obj->magic!=DMMAGIC)
  {
   dprintf(0,"Bad magic number at obj=%x (%d)\n",(int)obj,
                                                ((int)obj)-((int)firstobj));
  }


  freenode=obj->free;
  balance=obj->balance;

/*  if(obj!=tfind(obj->key,d)) dprintf(0,"Lost object obj=%x key=%d\n",(int)obj,obj->key); */

  dprintf(0,"object1");

  if(obj->LEFT!=NIL_NODE)
  {
   child=(objstr*)((char*)obj+obj->LEFT);

   if(child<firstobj || child>maxobj)
   {
    dprintf(0,"Bad left child at obj=%x child=%x %d\n",
           (int)obj,(int)child,((int)child)-((int)firstobj));
   }
   else
   {
    if(child->parent!=-obj->LEFT)
    {
     dprintf(0,"Bad left offset at obj=%x child=%x %d %d\n",(int)obj,(int)child,child->parent,obj->LEFT);

     dprintf(0,"flags=%x copyclass=%x balance=%x",child->flags,child->copyclass,child->balance);
     dprintf(0,"free=%x count=%x length=%x type=%d key=%x",child->free,child->count,child->length,child->type,child->key);
     dprintf(0,"left=%x right=%x",child->sibls[0],child->sibls[1]);

    }
    if(child->magic!=DMMAGIC)
    {
     dprintf(0,"Bad magic child at obj=%x child=%x %d\n",
                (int)obj,(int)child,((int)child)-((int)firstobj));
    }
    freenode-=child->free;
    leftb=child->balance;
   }
  }
  else leftb=1;

  dprintf(0,"object2");

  if(obj->RIGHT!=NIL_NODE)
  {
   dprintf(0,"object2a");

   child=(objstr*)((char*)obj+obj->RIGHT);

   if(child<firstobj || child>maxobj)
   {
    dprintf(0,"Bad right child at obj=%x child=%x %d\n",
           (int)obj,(int)child,((int)child)-((int)firstobj));
   }
   else
   {
    if(child->parent!=-obj->RIGHT)
    {
     dprintf(0,"Bad right offset at obj=%x child=%x %d %d\n",(int)obj,(int)child,child->parent,obj->RIGHT);

     dprintf(0,"flags=%x copyclass=%x balance=%x",child->flags,child->copyclass,child->balance);
     dprintf(0,"free=%x count=%x length=%x type=%d key=%x",child->free,child->count,child->length,child->type,child->key);
     dprintf(0,"left=%x right=%x",child->sibls[0],child->sibls[1]);


    }

    dprintf(0,"object2c");

    if(child->magic!=DMMAGIC)
    {
     dprintf(0,"Bad magic child at obj=%x child=%x %d\n",
               (int)obj,(int)child,((int)child)-((int)firstobj));
    }
    else
     dprintf(0,"Good magic child at obj=%x child=%x %d\n",(int)obj,(int)child,
                  ((int)child)-((int)firstobj));

    dprintf(0,"object2d");


    freenode-=child->free;
    rightb=child->balance;
   }
  }
  else rightb=-1;

 dprintf(0,"Object3");

  if(freenode>obj->length)
  {
   dprintf(0,"Object has more free than length at obj=%x key=%d\n",(int)obj,obj->key);
  }

/*  if(freenode!=obj->count)
  {
   printf("Object free!=count at obj=%x key=%d free=%d count=%d\n",
                                (int)obj,obj->key,freenode,obj->count);
  }
 */

 dprintf(0,"Object4");

  if(
     (obj->LEFT  && !obj->RIGHT && obj->balance!=1) ||
     (obj->RIGHT && !obj->LEFT && obj->balance!=-1))
  {
   dprintf(0,"bad balance at obj=%x key=%d\n",(int)obj,obj->key);
   dprintf(0,"left=%d right=%d balance=%d\n",obj->LEFT,obj->RIGHT,obj->balance);
  }

  freetotal+=freenode;

  dprintf(0,"freenode %d",freenode);

  lastobj=obj;

  obj=(objstr*)((char*)obj+obj->length);

  if(obj>=maxobj) break;
 }
 }

 if(obj>maxobj) 
              dprintf(0,"object pointed past end of domain at obj=%x\n",(int)obj);

 if(lastobj->key!=d->maxobject) dprintf(0,"max objects wrong %d %d\n",lastobj->key,d->maxobject);


 root=(objstr*)(d->domaindata+d->root);
 if(root->free!=freetotal)
 dprintf(0,"disagreement of free space %d %d",root->free,freetotal);
}



#endif






#ifdef MEMTEST



static void ob_test(void)
{
 domainstr * d;
 int         n;
 objstr    * obj;
 objhdrstr * sobj;
 int         i;
 int         j;
 int         free;
 int         len;
 int         max;

 ob_createdomain(&d);


 for(i=0x100;i<=0x2FF;i++)
 {
  ob_create(&n,128,d,0);
 }

 max=0;

 for(j=0;j<100000;j++)
 {
  ob_testtree(d);

  obj=tfind(rand() % d->maxobject,d);
  if(obj)
  {
   max++;
   ob_delete(obj->key,d);
   printf("deleted %x %d\n",obj->key,max);
  }
  else
  {
   n=3*sizeof(int)+(rand() & 0xFF);
   ob_create(&i,n,d,123);
   printf("created %x\n",i);
  }
 }
}




int main(int argc,char * argv[])
{
 USE(argc);
 USE(argv);

 ob_test();

 return(0);
}

#endif


