/*->c.object */

#include <stdafx.h>

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "transform.h"

#include "err.h"
#include "fsx.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"

#include "etc.h"

#include "constants.h"
#include "str.h"


#include "file.h"

#include "object.h"


#include "viewr.h"


typedef struct objecttype
{
 int         type;
 objectstr * ptr;

} objecttype;


#define OBCHUNK 16*sizeof(objecttype)


static objecttype * table;
static int          tablesize;



static int cmpobjects(const void * object1,const void * object2)
{
 return(((objecttype*)object1)->type-((objecttype*)object2)->type);
}


static objectstr * findobject(int type)
{
 objecttype * typeptr;
 objecttype   target;

 target.type=type;

 typeptr=(objecttype*)bsearch(&target,table,tablesize,sizeof(objecttype),cmpobjects);

 if(typeptr) return(typeptr->ptr);
 else        return(NULL);
}


os_error * objectcopy(objectloc * src,objectloc * dest,transformstr * tr)
{
 os_error  * err;
 objectstr * ptr;
 objhdrstr * objhdr;

 err=NULL;

 objhdr=(objhdrstr*)fob_find(src->object,src->file);
 if(objhdr)
 {
  ptr=findobject(objhdr->type);
  if(ptr && ptr->objectcopy) err=ptr->objectcopy(src,dest,tr);
 }
 return(err);
}


os_error * objectredraw(viewredrawstr * v,objhdrstr * objhdr)
{
 os_error  * err;
 objectstr * ptr;

 err=NULL;

 if(objhdr)
 {
  ptr=findobject(objhdr->type);
  if(ptr && ptr->objectredraw) err=ptr->objectredraw(v,objhdr);
 }
 return(err);
}


os_error * objectnew(objectloc * loc,int type,void * data)
{
 os_error  * err;
 objectstr * ptr;
 ptr=findobject(type);
 if(ptr && ptr->objectnew) err=ptr->objectnew(loc,data);
 else                      err=NULL;
 return(err);
}


os_error * objectcreate(objectloc * loc,int type,curvestr * c)
{
 os_error  * err;
 objectstr * ptr;
 ptr=findobject(type);
 if(ptr && ptr->objectcreate) err=ptr->objectcreate(loc,c);
 else                         err=NULL;
 return(err);
}


int objectfind(posnstr * posn,objhdrstr * objhdr,int * before,int last)
{
 objectstr * ptr;
 int         found;

 if(objhdr->key==*before) *before=0;
 else
 if(*before)
 {
  ptr=findobject(objhdr->type);
  found=ptr->objectfind(posn,objhdr,before,last);
  if(found)
  {
   if(found==*before) *before=0;
   else               last=found;
  }
 }
 return(last);
}



int objectselect(objhdrstr * objhdr,posnstr * posn,int type)
{
 objectstr * ptr;

 ptr=findobject(objhdr->type);
 if(ptr->objectselect) type=ptr->objectselect(objhdr,posn,type);

 return(type);
}


os_error * objectmessage(objhdrstr * objhdr,filestr * file,int * message,
                                                                 void * data)
{
 os_error  * err;
 objectstr * ptr;

 ptr=findobject(objhdr->type);
 if(ptr->objectmessage) err=ptr->objectmessage(objhdr,file,message,data);
 else                   err=NULL;

 return(err);
}






os_error * objectmessage2(int object,filestr * file,int * message,void * data)
{
 if(object) return(objectmessage((objhdrstr*)fob_find(object,file),file,message,data));
 else       return(NULL);
}




/* send this message to every element in a chain */

/* fundamental problem that a message can't make an object vanish */

os_error * objectchainmessage(int next,filestr * file,
                              int * message,void * data)
{
 os_error      * err;
 chainobject   * chain;

 err=NULL;

 while(next && (*message) && !err)
 {
  chain=(chainobject*)fob_find(next,file);

 /* next=chain->chain.next;  !!!! next may vanish after message */
 /* e.g. uncat page on style map change                         */

  err=objectmessage((objhdrstr*)chain,file,message,data);

  chain=(chainobject*)fob_find(next,file);
  if(chain) next=chain->chain.next;
  else      break;
 }

 return(err);
}


/* sends a message to all sections, and so hopefully to all objects */

os_error * broadcastmessage(filestr * file,int * message,void * data)
{
 os_error      * err;
 int             doc;
 int             mst;

 doc=file->hdr.section[DOC];
 mst=file->hdr.section[MST];

                        err=objectchainmessage(doc,file,message,data);
 if(!err && (*message)) err=objectchainmessage(mst,file,message,data);

 return(err);
}




os_error * broadcastmessageb(filestr * file,int branch,int * message,
                                                       void * data)
{
 os_error      * err;

 err=objectchainmessage(file->hdr.section[branch],file,message,data);

 return(err);
}



os_error * objectdb(objhdrstr * objhdr,filestr * file,
                                       viewstr * view,void * data)
{
 os_error  * err;
 objectstr * ptr;

 ptr=findobject(objhdr->type);
 if(ptr->objectdb) err=ptr->objectdb(objhdr,file,view,data);
 else              err=NULL;

 return(err);
}


os_error * objecttransform(objhdrstr * objhdr,filestr * file,
                              transformfn trf,int flags)
{
 os_error  * err;
 objectstr * ptr;

 err=NULL;

 if(objhdr)
 {
  ptr=findobject(objhdr->type);
  if(ptr && ptr->objecttransform)  
                       err=ptr->objecttransform(objhdr,file,trf,flags);
 }
 return(err);
}



os_error * objectdelete(objhdrstr * objhdr,int dm,filestr * file)
{
 os_error  * err;
 objectstr * ptr;

 err=NULL;

 if(objhdr)
 {
  ptr=findobject(objhdr->type);
  if(ptr && ptr->objectdelete) err=ptr->objectdelete(objhdr,dm,file);
 }
 return(err);
}




os_error * objectflow(objhdrstr * objhdr,filestr * file,objectflowstr * flow,
                                                                     int read)
{
 os_error  * err;
 objectstr * ptr;

 err=NULL;

 if(objhdr)
 {
  ptr=findobject(objhdr->type);
  if(ptr && ptr->objectflow) err=ptr->objectflow(objhdr,file,flow,read);
 }
 return(err);
}



os_error * objectsetflow(objhdrstr * start,objhdrstr * fix,
                                      filestr * file,int * end)
{
 os_error  * err;
 objectstr * ptr;

 err=NULL;

 if(start)
 {
  ptr=findobject(start->type);
  if(ptr && ptr->objectsetflow) err=ptr->objectsetflow(start,fix,file,end);
 }
 return(err);
}



os_error * objectmaster(int object,filestr * file,int * master,
                                                  objectbitstr * bits)
{
 os_error  * err;
 objectstr * ptr;
 objhdrstr * obj;

 err=NULL;

 obj=(objhdrstr*)fob_find(object,file);

 if(obj)
 {
  ptr=findobject(obj->type);
  if(ptr && ptr->objectmaster) err=ptr->objectmaster(obj,file,master,bits);
 }

 return(err);
}



os_error * objectlocal(int object,filestr * file,int makelocal)
{
 os_error  * err;
 objectstr * ptr;
 objhdrstr * obj;

 err=NULL;

 obj=(objhdrstr*)fob_find(object,file);

 if(obj)
 {
  ptr=findobject(obj->type);
  if(ptr && ptr->objectlocal) err=ptr->objectlocal(obj,file,makelocal);
 }

 return(err);
}




os_error * objectreflow(objhdrstr * objhdr,filestr * file)
{
 os_error  * err;
 objectstr * ptr;

 err=NULL;

 if(objhdr)
 {
  ptr=findobject(objhdr->type);
  if(ptr && ptr->objectreflow) err=ptr->objectreflow(objhdr,file);
 }

 return(err);
}



os_error * objectloadtype(char * name,int type,posnstr * posn,int * method)
{
 os_error  * err;
 objectstr * ptr;
 ptr=findobject(posn->otype);
 if(ptr && ptr->objectloadtype) err=ptr->objectloadtype(name,type,posn,method);
 else                           err=NULL;
 return(err);
}


os_error * objectloadpost(posnstr * posn,int flags)
{
 os_error  * err;
 objectstr * ptr;
 ptr=findobject(posn->otype);
 if(ptr && ptr->objectloadpost) err=ptr->objectloadpost(posn,flags);
 else                           err=NULL;
 return(err);
}


os_error * objectload(char * name,int type,int xvolatile,posnstr * posn)
{
 os_error  * err;
 objectstr * ptr;
 ptr=findobject(posn->otype);
 if(ptr && ptr->objectload) err=ptr->objectload(name,type,xvolatile,posn);
 else                       err=NULL;
 return(err);
}



os_error * objectshape(int object,int shape,int data,filestr * file)
{
 os_error  * err;
 objectstr * ptr;
 objhdrstr * obj;

 err=NULL;

 obj=(objhdrstr*)fob_find(object,file);

 if(obj)
 {
  ptr=findobject(obj->type);
  if(ptr && ptr->objectshape) err=ptr->objectshape(obj,shape,data,file);
 }

 return(err);
}



os_error * objectread(int action,objhdrstr * obj,
                                             filestr * file,void * data)
{
 os_error  * err;
 objectstr * ptr;

 err=NULL;

 if(obj)
 {
  ptr=findobject(obj->type);
  if(ptr && ptr->objectread) err=ptr->objectread(action,obj,file,data);
 }

 return(err);
}


os_error * objectoread(int action,int object,filestr * file,void * data)
{
 objhdrstr * obj;
 obj=(objhdrstr*)fob_find(object,file);
 return(objectread(action,obj,file,data));
}



void posntoloc(posnstr * posn,objectloc * loc)
{
 memset(loc,0,sizeof(objectloc));

 loc->file=posn->file;
 loc->object=posn->object;
 loc->section=posn->section;
 loc->page=posn->page;
 loc->branch=(posn->view)->branch;
}




os_error * objectregister(objectstr * object)
{
 os_error * err;
 int        i;
 int        j;

 if(table) 
    err=flex_chunk((flex_ptr)&table,(tablesize+1)*sizeof(objecttype),OBCHUNK);
 else
    err=flex_alloc((flex_ptr)&table,OBCHUNK);

 if(!err)
 {
  for(i=0;i<tablesize;i++) if(table[i].type>object->type) break;
  for(j=tablesize-1;j>=i;j--) table[j+1]=table[j];
  table[i].ptr=object;
  table[i].type=object->type;
  tablesize++;
 }

 return(err);
}

