/*->h.object */



typedef struct objhdrstr
{
 int key;
 int type;
} objhdrstr;



typedef struct sobjectstr
{
 int            key;
 int            type;
 chainstr       chain;
 flagstr        flags;

 boxstr         box;
 transformstr   transform;

 curvestr       fcurve;    /* frame curve    */
 curvestr       icurve;    /* internal curve */
 curvestr       ocurve;    /* external curve */

} sobjectstr;



typedef int (*transformfn)(transformstr * r,transformstr * t,
                                    curvestr * c,transformstr * ct);




#define LOCNEW         1
#define LOCFLOWDATA    2             /* we copy vboxes */
#define LOCUNDO        4
#define LOCNVBOX       8             /* on new don't create vboxes */
#define LOCOVERWRITE   0x10          /* styles, colours            */
#define LOCNOUPDATE    0x20          /* no flows on dscript create */
#define LOCMOVE        0x40
#define LOCCLIP        0x80          /* destination is clipboard */
/* #define LOCPASTE       0x100       not used paste from clipboard     */
#define LOCOLDMASTER   0x200         /* copy from old masters    */
#define LOCNOFLOW      0x400         /* no reflows for other objects */
/* #define LOCREALCOPY    0x800       not used               */
#define LOCKEEPMASTER  0x1000        /* doc to doc copy, map maste refs. */



typedef struct objectloc
{
 filestr * file;
 int       object;
 int       section;
 int       page;
 int       parent;
 int       after;
 int       chain;
 int       branch;
 int       copymode;

 int       u;
 int       offset;

 int       spare[4];

} objectloc;





typedef struct objectflowstr
{
 textflowstr  flow;
 transformstr tr;
} objectflowstr;





#define DELETESCRAP   0x1
#define DELETEUNLINK  0x2


#define TRNOREFLOW  0x1
#define TRNOUPDATE  0x2
#define TRNOUNDO    0x4
#define TRNOREFLOWS 0x8
#define TRNOREFLOWD 0x10
#define TRADJUST    0x20
#define TRMIRROR    0x40



#define OBDIRECTLOAD 0x1
#define OBCARETLOAD  0x2


typedef os_error * (*objectnewfn)(objectloc * loc,void * data);
typedef os_error * (*objectcopyfn)(objectloc * src,objectloc * dest,
                                                         transformstr * tr);
typedef os_error * (*objectredrawfn)(viewredrawstr * v,objhdrstr * objhdr);
typedef os_error * (*objectcreatefn)(objectloc * loc,curvestr * c);
typedef int        (*objectfindfn)(posnstr *posn,objhdrstr *objhdr,
                                                     int * before,int last);
typedef int    (*objectselectfn)(objhdrstr * objhdr,posnstr * file,int type);

typedef os_error * (*objecttransformfn)(objhdrstr * objhdr,filestr * file,
                                           transformfn trf,int flags);

typedef os_error * (*objectflowfn)(objhdrstr * objhdr,filestr * file,
                                                objectflowstr * flow,int read);

typedef os_error * (*objectloadtypefn)(char * name,int type,posnstr * posn,int * method);
typedef os_error * (*objectloadpostfn)(posnstr * posn,int flags);
typedef os_error * (*objectloadfn)(char * name,int type,int xv,posnstr * posn);

typedef os_error * (*objectmessagefn)(objhdrstr * objhdr,filestr * file,
                                                  int * message,void * data);
typedef os_error * (*objectdbfn)(objhdrstr * objhdr,filestr * file,
                                                   viewstr * view,void * data);
typedef os_error * (*objectdeletefn)(objhdrstr * objhdr,int dm,filestr * file);

typedef os_error * (*objectreflowfn)(objhdrstr * objhdr,filestr * file);

typedef os_error * (*objectsetflowfn)(objhdrstr * start,objhdrstr * fix,
                                      filestr * file,int * end);

typedef os_error * (*objectmasterfn)(objhdrstr * obj,filestr * file,
                                         int * master,objectbitstr * bits);

typedef os_error * (*objectshapefn)(objhdrstr * obj,int shape,int data,
                                                              filestr * file);

typedef os_error * (*objectreadfn)(int action,objhdrstr * obj,
                                             filestr * file,void * data);



#define MAKELOCAL    1
#define MAKEMASTER   0
#define FORCEMASTER  2
#define FORCEMASTER1 3
#define FORCEMASTERU 4
#define FORCEMASTERR 5
#define DELETEMASTER 6



typedef os_error * (*objectlocalfn)(objhdrstr * obj,filestr * file,
                                                               int makelocal);



typedef struct objectstr
{
 int               type;
 objectnewfn       objectnew;
 objectcopyfn      objectcopy;
 objectredrawfn    objectredraw;
 objectcreatefn    objectcreate;
 objectfindfn      objectfind;
 objectselectfn    objectselect;
 objecttransformfn objecttransform;
 objectloadtypefn  objectloadtype;
 objectloadfn      objectload;
 objectloadpostfn  objectloadpost;
 objectmessagefn   objectmessage;
 objectdbfn        objectdb;
 objectflowfn      objectflow;
 objectdeletefn    objectdelete;
 objectreflowfn    objectreflow;
 objectsetflowfn   objectsetflow;
 objectmasterfn    objectmaster;
 objectlocalfn     objectlocal;
 objectshapefn     objectshape;
 objectreadfn      objectread;
 char          *   x12;
} objectstr;




extern os_error * objectregister(objectstr * object);

extern os_error * objectcopy(objectloc * src,objectloc * dest,
                                                        transformstr * tr);
extern os_error * objectnew(objectloc * loc,int type,void * data);
extern os_error * objectredraw(viewredrawstr * v,objhdrstr * objhdr);
extern os_error * objectcreate(objectloc * loc,int type,curvestr * c);
extern int        objectfind(posnstr *posn,objhdrstr * objhdr,
                                               int * before,int last);
extern int        objectselect(objhdrstr * objhdr,posnstr * posn,int type);
extern os_error * objectmessage(objhdrstr * objhdr,filestr * file,
                                     int * message,void * data);

extern os_error * objectmessage2(int object,filestr * file,
                                     int * message,void * data);

extern os_error * objectchainmessage(int object,filestr * file,
                                                  int * message,void * data);

extern os_error * broadcastmessage(filestr * file,int * message,void * data);
extern os_error * broadcastmessageb(filestr * file,int branch,int * message,
                                                       void * data);


extern os_error * objecttransform(objhdrstr * objhdr,filestr * file,
                                     transformfn trf,int flags);

extern os_error * objectloadtype(char * name,int type,posnstr * posn,int * method);
extern os_error * objectloadpost(posnstr * posn,int flags);
extern os_error * objectload(char * name,int type,int xv,posnstr * posn);

extern os_error * objectdb(objhdrstr * objhdr,filestr * file,
                                              viewstr * view,void * data);
extern os_error * objectflow(objhdrstr * objhdr,filestr * file,
                                                objectflowstr * flow,int read);

extern os_error * objectdelete(objhdrstr * objhdr,int dm,filestr * file);

extern os_error * objectreflow(objhdrstr * objhdr,filestr * file);

extern os_error * objectsetflow(objhdrstr * start,objhdrstr * fix,
                                      filestr * file,int * end);

extern os_error * objectmaster(int object,filestr * file,int * master,
                                                         objectbitstr * bits);

extern os_error * objectlocal(int object,filestr * file,int makelocal);

extern os_error * objectshape(int object,int shape,int data,filestr * file);




#define SELECTINFO 0x1

typedef struct outlineinfostr
{
 transformstr tr;
 curvestr     fcurve;
 int          frame;

} outlineinfostr;


#define OREADFILL 0x2


extern os_error * objectread(int action,objhdrstr * objhdr,
                                        filestr * file,void * data);

extern os_error * objectoread(int action,int object,
                                         filestr * file,void * data);


extern void       posntoloc(posnstr * posn,objectloc * loc);


/*************************************************************************/

typedef struct linkstr
{
 int dprev;
 int dpage;
 int dnext;
 int dsection;

 int mprev;
 int mpage;
 int mnext;
 int msection;
} linkstr;



typedef struct mlinkstr
{
 filestr * sfile;
 int       spage[2];

 filestr * dfile;
 int       dpage[2];

 int       hand;

} mlinkstr;



typedef struct dlinkstr
{
 filestr * sfile;
 filestr * dfile;

} dlinkstr;



typedef os_error * (*omessagefn)(objhdrstr * objhdr,filestr * file,
                                                    int * message);


typedef void       (*snapfn)(int object,transformstr * tr,boxstr * sbox);


typedef struct plinkdatastr
{
 int nf;
 int nl;

} plinkdatastr;

