/*->c.link */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>  

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "bbc.h"
#include "akbd.h"
#include "transform.h"

#include "err.h"
#include "fsx.h"
#include "xext.h"
#include "poll.h"
#include "alloc.h"
#include "trans.h"
#include "etc.h"
#include "colour.h"
#include "key.h"
#include "xmath.h"
#include "conf.h"
#include "dbhi.h"

#include "str.h"
#include "constants.h"

#include "mhi.h"
#include "object.h"
#include "chain.h"
#include "reslink.h"


#include "viewr.h"
#include "view.h"
#include "file.h"
#include "caret.h"
#include "select.h"
#include "posn.h"
#include "xredraw.h"
#include "xrline.h"
#include "curve.h"
#include "undo.h"
#include "frameflow.h"

#include "toolstr.h"
#include "tools.h"
#include "toolbox.h"


#include "txstr.h"

#include "codes.h"
#include "vl.h"
#include "tx.h"
#include "vec.h"
#include "text.h"
#include "mx.h"
#include "fl.h"


#include "link.h"



#define LI00 0x500
#define LI01 0x501
#define LI02 0x502
#define LI03 0x503
#define LI04 0x504
#define LI05 0x505
#define LI06 0x506
#define LI07 0x507
#define LI08 0x508
#define LI09 0x509
#define LI11 0x50B
#define LI12 0x50C
#define LI13 0x50D
#define LI14 0x50E
#define LI15 0x50F
#define LI16 0x510
#define LI17 0x511


/*

LI00:Cannot make links from autolinked chains
LI01:You can only link text frames!
LI02:Cannot link into autolinked chains
LI03:You can only link to another text frame!
LI04:Cannot link frames in different chapters!
LI05:Autolink master page frames must be empty
LI06:You cannot link a frame with itself!
LI07:You cannot link to a frame that is already linked!
LI08:Cannot have recursive or repetitive links!
LI09:Cannot link two frames that both contain text!
LI10:Link frame
LI11:Cannot break links between master frames!
LI12:Cannot break links between autolinked frames - remove autolink first!
LI13:Cannot link embedded frames!
LI14:Cannot link to master frames in document!
LI15:Cannot link to grouped frames!
LI16:Autolink must be from end of chain to start.
LI17:Link must be from left page to right.

*/






#define ULINK    0x6200
#define UDELINK  0x6201
#define USLINK   0x6202         
#define USCLOSE  0x6203
#define USOPEN   0x6204




static pointstr linkpointer=
{
 "p_link",0,0
};



typedef struct linktoolstr
{
 int linkmode;
 int linkbox;
 int linksection;
} linktoolstr;




static int linkdrag;
static int xlinkbox;      /* link box not to redraw */

static posnstr start;
static posnstr end;



static void refreshlink(vector * vb1,vector * vb2,int layer);
static void setlinkbox(filestr * file,linktoolstr * ltool);
static void clearlinkbox(filestr * file,linktoolstr * ltool);
static os_error * cutlink(void);





#define LINKTHICK 2835

#define ECWIDTH  2
#define ECHEIGHT 5



static void genline(vector * vb1,vector * vb2,xlinestr * r,viewstr * view)
{
 textboxstr * textp;
 filestr    * file;
 int          xpo1;
 int          ypo1;
 int          xpo2;
 int          ypo2;


 file=view->file;

 getpagecoords(view,vb1->page,vb1->section,&xpo1,&ypo1);
 getpagecoords(view,vb2->page,vb2->section,&xpo2,&ypo2);

 r->thickness=LINKTHICK;
 r->endcaps=2;
 r->style=0;
 r->ecw=ECWIDTH;
 r->ech=ECHEIGHT;

 textp=(textboxstr*)fob_find(vb1->textsrc.text,file);

 r->ex=xpo1+textp->box.x+textp->box.w-(textp->box.w>>3);
 r->ey=ypo1+textp->box.y+textp->box.h-(textp->box.h>>3);

 textp=(textboxstr*)fob_find(vb2->textsrc.text,file);

 r->ox=xpo2+textp->box.x+(textp->box.w>>3);
 r->oy=ypo2+textp->box.y+(textp->box.h>>3);

 setxlineboxoffends(r);
}









/*****************************************************************************/

typedef struct linkblockstr
{
 int box1;
 int box2;
 int src1;
 int vsrc1;
 int src2;
 int vsrc2;
 int section;

} linkblockstr;





static os_error * lsaveposn(filestr * file)
{
 os_error    * err;
 int           temp[2];
 toolboxstr  * toolbox;
 linktoolstr * ltool;

 toolbox=file->toolbox;
 ltool=(linktoolstr*)(toolbox->toolstate[toolindex(TL_LINK)].hprivate);

 temp[0]=ltool->linkbox;
 temp[1]=ltool->linksection;

 err=undowritefn(file,USLINK,temp,2*sizeof(int));

 return(err);
}



static os_error * slink(filestr * file,int box,int section)
{
 os_error    * err;
 toolboxstr  * toolbox;
 linktoolstr * ltool;

 toolbox=file->toolbox;
 ltool=(linktoolstr*)(toolbox->toolstate[toolindex(TL_LINK)].hprivate);

 err=lsaveposn(file);

 if(ltool->linkbox) clearlinkbox(file,ltool);
 ltool->linkbox=box;
 ltool->linksection=section;
 if(ltool->linkbox) setlinkbox(file,ltool);

 return(err);
}


static os_error * unslink(filestr * file)
{
 int temp[2];
 undoread(file,temp,2*sizeof(int));
 return(slink(file,temp[0],temp[1]));
}





static os_error * linkdelink(linkblockstr * linkblock,filestr * file)
{
 os_error    * err;
 vector        vb1;
 vector        vb2;
 int           box1;
 int           box2;
 textboxstr  * textp;
 int           text;
 texsrc      * texsp;
 toolboxstr  * toolbox;
 linktoolstr * ltool;
 flowstr       flop;
 int           fstart;


 err=NULL;


 box1=linkblock->box1;
 box2=linkblock->box2;

 setpagevectorfrombox(&vb1,box1,file);
 setvectorfrombox(&vb2,0,box2,file); /* later we will use the offset */

 xlinkbox=box1;
 refreshlink(&vb1,&vb2,VRBACK);
 xlinkbox=0;

 toolbox=file->toolbox;
 ltool=(linktoolstr*)(toolbox->toolstate[toolindex(TL_LINK)].hprivate);
 ltool->linkbox=0;


 text=box1;
 fob_mod(text,file,(fobjptr)&textp);
 textp->fchain.next=0;

 while(1)
 {
  fstart=text;

  if(linkblock->src1)
  {
   texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
   texsp->source=linkblock->src1;
   texsp->vsource=linkblock->vsrc1;
   zerocolumns(textp,0,0);
   viewrefreshobject(text,0,0,file);
  }

  text=textp->fchain.prev;
  if(!text) break;
  fob_mod(text,file,(fobjptr)&textp);
 }


 if(linkblock->src1)
 {
  setvectorfrombox(&vb1,0,fstart,file);

  flop.opcode=FWAIT;
  flop.data.type=FLDELTA;
  flop.data.offset=0;
  flop.data.length=0;
  flop.data.delta=sourcesize(vb1.textsrc.source,file);
  flow(&vb1,&flop);
 }
 else
 {
  if(sourcesize(vb1.textsrc.source,file)>vb2.boffset)
  {
   textp->flags.full=1;
   refreshfullbox(box1,file);
  }
 }

 fstart=text=box2;
 fob_mod(text,file,(fobjptr)&textp);
 textp->fchain.prev=0;

 while(1)
 {
  if(linkblock->src2)
  {
   texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
   texsp->source=linkblock->src2;
   texsp->vsource=linkblock->vsrc2;
   zerocolumns(textp,0,0);
   viewrefreshobject(text,0,0,file);
  }

  text=textp->fchain.next;
  if(!text) break;
  fob_mod(text,file,(fobjptr)&textp);
 }

 setvectorfrombox(&vb1,0,fstart,file);

 flop.opcode=FWAIT;
 flop.data.type=FLDELTA;
 flop.data.offset=0;
 flop.data.length=0;
 flop.data.delta=sourcesize(linkblock->src2,file);
 flow(&vb1,&flop);

 return(err);
}






static os_error * linklink(linkblockstr * linkblock,filestr * file)
{
 os_error    * err;
 int           src1;
 int           vsrc1;
 int           src2;
 int           vsrc2;
 int           fstart; /* first box in the chain */
 textboxstr  * textp;
 int           box;
 texsrc      * texsp;
 flowstr       flop;
 vector        v;
 vector        vb1;
 vector        vb2;
 int           box1;
 int           box2;


 err=NULL;


 box1=linkblock->box1;
 box2=linkblock->box2;


 textp=(textboxstr*)fob_find(box1,file);
 texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
 src1=texsp->source;
 vsrc1=texsp->vsource;


 /* if the first box has a full marker then it should be cleared */

 if(textp->flags.full)
 {
  textp->flags.full=0;
  refreshfullbox(box1,file);
 }


 textp=(textboxstr*)fob_find(box2,file);
 texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
 src2=texsp->source;
 vsrc2=texsp->vsource;


 if(!nullsource(src2,file))
 {
  /* zap source 1 */

  linkblock->src1=src1;
  linkblock->vsrc1=vsrc1;

  fob_mod(box2,file,(fobjptr)&textp);
  textp->fchain.prev=box1;
  fob_mod(box1,file,(fobjptr)&textp);
  textp->fchain.next=box2;
  box=box1;
  while(1)
  {
   fstart=box;
   texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
   texsp->source=src2;
   texsp->vsource=vsrc2;
   box=textp->fchain.prev;
   if(!box) break;
   fob_mod(box,file,(fobjptr)&textp);
  }

  box=fstart;
  while(box!=box2)     /* alternative is to zero the whole chain */
  {
   fob_mod(box,file,(fobjptr)&textp);
   zerocolumns(textp,0,0);
   box=textp->fchain.next;
  }

 /* vsourcezero(vsrc2,file); */

  /* reflow from the begining */

  setvectorfrombox(&v,0,fstart,file);

  flop.opcode=FWAIT;
  flop.data.type=FLDIRTY; /* then do a delta reflow */
  flop.data.offset=0;   /* leads to old lines not being removed from screen */
  flop.data.length=sourcesize(src2,file);
  flop.data.delta=0;

  flow(&v,&flop);
 }
 else
 {
  /* zap source 2 */

  linkblock->src2=src2;
  linkblock->vsrc2=vsrc2;

  fob_mod(box1,file,(fobjptr)&textp);
  textp->fchain.next=box2;
  fob_mod(box2,file,(fobjptr)&textp);
  textp->fchain.prev=box1;

  box=box2;
  while(box)
  {
   fob_mod(box,file,(fobjptr)&textp);
   zerocolumns(textp,0,0);

   texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
   texsp->source=src1;
   texsp->vsource=vsrc1;

   box=textp->fchain.next;
  }

  /* reflow from box2 */

  setvectorfrombox(&v,0,box2,file);

  flop.opcode=FWAIT;
  flop.data.type=FLDELTA;
  flop.data.offset=0;
  flop.data.length=0;
  flop.data.delta=sourcesize(src1,file)-v.boffset;

  flow(&v,&flop);
 }

 setpagevectorfrombox(&vb1,box1,file);
 setpagevectorfrombox(&vb2,box2,file);

 refreshlink(&vb1,&vb2,VRTOOLS);

 return(err);
}




static os_error * unlink(filestr * file)
{
 linkblockstr linkblock;


 undoread(file,&linkblock,sizeof(linkblockstr));
 undowritefn(file,UDELINK,&linkblock,sizeof(linkblockstr));
 linklink(&linkblock,file);

 return(NULL);
}




static os_error * undelinkscrap(filestr * file)
{
 linkblockstr linkblock;

 undoread(file,&linkblock,sizeof(linkblockstr));

 txsourcedel(linkblock.src1,file,1);
 txsourcedel(linkblock.src2,file,1);
 vsourcedel(linkblock.vsrc1,file);
 vsourcedel(linkblock.vsrc2,file);

 return(NULL);
}



static os_error * undelink(filestr * file)
{
 linkblockstr linkblock;

 undoread(file,&linkblock,sizeof(linkblockstr));
 undowritefn(file,ULINK,&linkblock,sizeof(linkblockstr));
 linkdelink(&linkblock,file);

 return(NULL);
}





/**************************************************************************/



static os_error * linkclose(linkblockstr * linkblock,filestr * file)
{
 os_error    * err;
 textboxstr  * textp;
 vector        vb1;
 vector        vb2;
 int           box1;
 int           box2;
 int           text;


 err=NULL;

 text=box1=linkblock->box1;
 box2=linkblock->box2;

 while(text)
 {
  textp=(textboxstr*)fob_find(text,file);
  textp->data.repeat=1;
  text=textp->fchain.next;
 }

 setpagevectorfrombox(&vb1,box1,file);
 setpagevectorfrombox(&vb2,box2,file);

 refreshlink(&vb2,&vb1,VRTOOLS);

 return(err);
}







static os_error * linkopen(linkblockstr * linkblock,filestr * file)
{
 os_error    * err;
 textboxstr  * textp;
 vector        vb1;
 vector        vb2;
 int           box1;
 int           box2;
 int           text;


 err=NULL;

 text=box1=linkblock->box1;
 box2=linkblock->box2;

 setpagevectorfrombox(&vb1,box1,file);
 setpagevectorfrombox(&vb2,box2,file);

 xlinkbox=box2;
 refreshlink(&vb2,&vb1,VRBACK);
 xlinkbox=0;

 while(text)
 {
  textp=(textboxstr*)fob_find(text,file);
  textp->data.repeat=0;
  text=textp->fchain.next;
 }

 return(err);
}



static os_error * usclose(filestr * file)
{
 linkblockstr linkblock;

 undoread(file,&linkblock,sizeof(linkblockstr));
 undowritefn(file,USOPEN,&linkblock,sizeof(linkblockstr));
 linkclose(&linkblock,file);

 return(NULL);
}



static os_error * usopen(filestr * file)
{
 linkblockstr linkblock;

 undoread(file,&linkblock,sizeof(linkblockstr));
 undowritefn(file,USCLOSE,&linkblock,sizeof(linkblockstr));
 linkopen(&linkblock,file);

 return(NULL);
}




/**************************************************************************/




static os_error * removelink(int box1,int section,filestr * file)
{
 os_error      * err;
 linkblockstr    linkblock;
 linkblockstr    linkblock2;
 int             text;
 textboxstr    * textp;
 int             open;
 int             unlink;
 int             box2;
 int             page;
 int             page1;
 int             text1;


 err=NULL;

 open=0;
 unlink=1;

 textp=(textboxstr*)fob_find(box1,file);
 if(textp->master)       err=geterror(LI11);
 else
 if(textp->data.repeat)
 {
  if(textp->fchain.next) err=geterror(LI12);
  else
  {
   open=1;
   unlink=0;

   page1=pageof((chainobject*)textp,file);
   text1=text=box1;

   while((text=textp->fchain.prev)!=0)
   {
    textp=(textboxstr*)fob_find(text,file);
    page=pageof((chainobject*)textp,file);

    if(page!=page1)
    {
     memset(&linkblock2,0,sizeof(linkblockstr));
     unlink=1;
     linkblock2.box1=text;
     linkblock2.box2=text1;
     linkblock2.section=section;
     break;
    }

    page1=page;
    text1=text;
   }
  }
 }


 if(!err) err=modfile(file,section);
 if(!err) err=undostart(file);
 if(!err)
 {
  slink(file,0,section);

  if(open)
  {
   text=box2=box1;
   while(text)
   {
    box2=text;
    textp=(textboxstr*)fob_find(text,file);
    text=textp->fchain.prev;
   }

   linkblock.box1=box2;
   linkblock.box2=box1;
   linkblock.section=section;

   linkopen(&linkblock,file);
   undowritefn(file,USCLOSE,&linkblock,sizeof(linkblockstr));
  }

  if(unlink)
  {
   if(open && unlink)
   {
    linkblock=linkblock2;
   }
   else
   {
    textp=(textboxstr*)fob_find(box1,file);
    memset(&linkblock,0,sizeof(linkblockstr));
    linkblock.box1=box1;
    linkblock.box2=textp->fchain.next;
    linkblock.section=section;
   }

   newtextbuffers(&linkblock.src2,&linkblock.vsrc2,file);

   err=linkdelink(&linkblock,file);

   undowritefn(file,ULINK,&linkblock,sizeof(linkblockstr));
  }

  undoend(file);
 }
 return(err);
}





#define LINKBNOLINK 0x1
#define LINKBCLOSE  0x2



static os_error * linkbox2box(int box1,int box2,int section,
                                                    filestr * file,int bits)
{
 os_error      * err;
 linkblockstr    linkblock;
 textboxstr    * textp;


          err=modfile(file,section);
 if(!err) err=undostart(file);
 if(!err)
 {
 /* problem (1) the box we are linking from, already has an arrow going out of
    it. We must delete this link, creating a new source for the remaining boxes
    and reflowing them */

  if(!(bits & LINKBNOLINK))
  {
   textp=(textboxstr*)fob_find(box1,file);
   if(textp->fchain.next) err=removelink(box1,section,file);

   memset(&linkblock,0,sizeof(linkblockstr));

   linkblock.box1=box1;
   linkblock.box2=box2;
   linkblock.section=section;

   err=linklink(&linkblock,file);
   undowritefn(file,UDELINK,&linkblock,sizeof(linkblockstr));

   slink(file,box1,linkblock.section);
  }

  if(bits & LINKBCLOSE)
  {
   textp=(textboxstr*)fob_find(box1,file);
   while(textp->fchain.prev) textp=(textboxstr*)fob_find(textp->fchain.prev,file);
   box1=textp->key;
   textp=(textboxstr*)fob_find(box1,file);
   while(textp->fchain.next) textp=(textboxstr*)fob_find(textp->fchain.next,file);
   box2=textp->key;

   linkblock.box1=box1;
   linkblock.box2=box2;
   linkblock.section=section;

   err=linkclose(&linkblock,file);
   undowritefn(file,USOPEN,&linkblock,sizeof(linkblockstr));
  }

  undoend(file);
 }
 return(err);
}




/**************************************************************************/
/* Link tool */

static viewredrawstr * linkrv;
static linktoolstr   * ltool;
static int             linkpn;



static os_error * linkr(objhdrstr * obj,filestr * file,int * message)
{
 os_error   * err;
 textboxstr * textp;
 xlinestr     templine;
 int          fnext;
 int          textk;
 vector       vb1;
 vector       vb2;
 pagestr    * pagep;
 int          state1;
 int          state2;
 int          repeat;


 err=NULL;

 if(obj->type==OTEXTBOX)
 {
  textp=(textboxstr*)obj;
  textk=textp->key;
  fnext=textp->fchain.next;
  if(!fnext && textp->data.repeat && !textp->master)
  {
   repeat=1;
   fnext=textk;
   while(textp->fchain.prev)
   {
    fnext=textp->fchain.prev;
    textp=(textboxstr*)fob_find(fnext,file);
   }
  }
  else repeat=0;

  if(fnext && textk!=xlinkbox)
  {
   setpagevectorfrombox(&vb1,textk,file);
   setpagevectorfrombox(&vb2,fnext,file);

   pagep=(pagestr*)fob_find(vb1.page,vb1.file);
   state1=linkpn-pagep->data.number;

   pagep=(pagestr*)fob_find(vb2.page,vb2.file);
   state2=linkpn-pagep->data.number;

   if((state1==0) || 
      (state2==0) ||
      (state1<0 && state2>0) ||
      (state1>0 && state2<0)
     )
   {
    genline(&vb1,&vb2,&templine,linkrv->view);
    if(vb1.textsrc.text==ltool->linkbox) 
                       wimp_setcolour(repeat?WIMPRED:WIMPBLACK);
    else               wimp_setcolour(repeat?WIMPGREEN:WIMPGREY2);
    setxlinevars(&templine,3,linkrv);
   }
  }
 }

 return(err);

 USE(message);
}



/*

  The problem with redraw is caused by bbc_draw() being limited to 15bits
  One possible solution is to clip the redraws to pages


*/

static os_error * linkredraw(toolboxstr * toolbox,viewredrawstr * v)
{
 os_error    * err;
 int           message;
 linktoolstr * ltool;
 sectionstr  * sectionp;
 pagestr     * pagep;
 filestr     * file;

 file=v->file;

 ltool=(linktoolstr*)(toolbox->toolstate[toolbox->current].hprivate);

 err=NULL;

 pagep=(pagestr*)fob_find(v->page,file);
 linkpn=pagep->data.number;


 if(ltool->linkmode && !v->onelayer)
 {
  linkrv=v;
  message=OM_FUNCTION;
  sectionp=(sectionstr*)fob_find(v->section,file);
  err=objectmessage((objhdrstr*)sectionp,file,&message,(void*)linkr);
 }


 if(linkdrag)
 {
  if(v->view==start.view)
  {
   rseteorcol(WIMPWHITE,WIMPGREY2);

   bbc_move(v->ox+osunits(start.xpo+start.x,v->view),
            v->oy-osunits(start.ypo+start.y,v->view));

   bbc_draw(v->ox+osunits(end.xpo+end.x,v->view),
            v->oy-osunits(end.ypo+end.y,v->view));
  }
 }

 return(err);
}




static void refreshlink(vector * vb1,vector * vb2,int layer)
{
 xlinestr  templine;
 aboxstr   abox;
 viewstr * view;
 filestr * file;

 file=vb1->file;
 view=file->view;

 while(view)
 {
  genline(vb1,vb2,&templine,view);
  setxlinevars(&templine,0,NULL);

  abox.x0=templine.bx;
  abox.x1=templine.bx+templine.bw;
  abox.y1=templine.by;
  abox.y0=templine.by+templine.bh;

  viewrefreshaarea(view,&abox,layer);

  view=view->next;
 }
}



static void setlinkbox(filestr * file,linktoolstr * ltool)
{
 vector       vb1;
 vector       vb2;
 textboxstr * textp;
 int          temp;

 textp=(textboxstr*)fob_find(ltool->linkbox,file);
 temp=textp->fchain.next;
 if(!temp && textp->data.repeat)
 {
  temp=ltool->linkbox;
  while(textp->fchain.prev)
  {
   temp=textp->fchain.prev;
   textp=(textboxstr*)fob_find(temp,file);
  }
 }

 setpagevectorfrombox(&vb1,ltool->linkbox,file);
 setpagevectorfrombox(&vb2,temp,file);

 refreshlink(&vb1,&vb2,VRTOOLS);
}



static void clearlinkbox(filestr * file,linktoolstr * ltool)
{
 vector       vb1;
 vector       vb2;
 textboxstr * textp;
 int          temp;

 textp=(textboxstr*)fob_find(ltool->linkbox,file);
 temp=textp->fchain.next;
 if(!temp && textp->data.repeat)
 {
  temp=ltool->linkbox;
  while(textp->fchain.prev)
  {
   temp=textp->fchain.prev;
   textp=(textboxstr*)fob_find(temp,file);
  }
 }

 setpagevectorfrombox(&vb1,ltool->linkbox,file);
 setpagevectorfrombox(&vb2,temp,file);

 ltool->linkbox=0;

 refreshlink(&vb1,&vb2,VRTOOLS);
}


/***************************************************************************/

#define LINKDELICON 52

static iconmap lipmap[]=
{
 LINKDELICON,0,0,IP_TIPS,  /* delete          0 */

 0,0,0,0,  /* terminator */
};




static posnstr * lposn;

static os_error * linkfn(objhdrstr * obj,filestr * file,int * message)
{
 os_error   * err;
 textboxstr * textp;
 xlinestr     templine;
 vector       vb1;
 vector       vb2;
 int          fnext;
 int          textk;
 int          repeat;


 err=NULL;

 if(obj->type==OTEXTBOX)
 {
  textp=(textboxstr*)obj;
  fnext=textp->fchain.next;
  textk=textp->key;

  if(!fnext && textp->data.repeat)
  {
   repeat=1;
   fnext=textk;
   while(textp->fchain.prev)
   {
    fnext=textp->fchain.prev;
    textp=(textboxstr*)fob_find(fnext,file);
   }
  }
  else repeat=0;

  if(fnext)
  {
   setpagevectorfrombox(&vb1,textk,file);
   setpagevectorfrombox(&vb2,fnext,file);

   genline(&vb1,&vb2,&templine,lposn->view);

   if(insidexline(lposn->x+lposn->xpo,lposn->y+lposn->ypo,&templine,
                                                             lposn->view)) 
   {
    /* make this the selected link */

    if(vb1.textsrc.text!=ltool->linkbox)
    {
     err=undostart(file);
     if(!err)
     {
      slink(file,vb1.textsrc.text,lposn->section);
      undoend(file);
     }
     *message=0;
    }
   }
  }
 }

 return(err);
}


static os_error * linkclicks(posnstr * posn)
{
 int branch;
 int message;

 lposn=posn;
 branch=posn->view->branch;
 message=OM_FUNCTION;
 return(broadcastmessageb(posn->file,branch,&message,(void*)linkfn));
}



static os_error * linkclick(toolboxstr * toolbox,posnstr * posn)
{
 os_error   * err;
 textboxstr * textp;

 err=NULL;

 ltool=(linktoolstr*)(toolbox->toolstate[toolbox->current].hprivate);

 if(posn->buttons==VBDRAGS || posn->buttons==VBLONGCLICKS)
 {
  if(findbox(posn,-1))
  {
   if(posn->object)
   {
    if(posn->otype!=OTEXTBOX) err=geterror(LI01);
    else
    if(embededobject(posn->object,posn->file)) err=geterror(LI13);
    else
    {
     textp=(textboxstr*)fob_find(posn->object,posn->file);
     if(textp->data.repeat) err=geterror(LI00);
     else
     if(!posn->view->branch && textp->master) err=geterror(LI14);
     else
     {
      if(childobject((chainobject*)textp,posn->file)) err=geterror(LI15);
      else
      {
       toolsstartdrag(posn,TOOLDRAGWINDOW);
       end=start=*posn;
       linkdrag=1;
      }
     }
    }
   }
  }
 }
 else
 if(posn->buttons==VBCLICKS)
 {
  err=linkclicks(posn);
 }

 caretensure(posn->view);

 return(err);
}


static os_error * linkdragend(toolboxstr * toolbox,posnstr * posn,
                                                             wimp_box * box)
{
 os_error   * err;
 textboxstr * textp;
 int          src2;
 int          src1;
 texsrc     * texsp;
 int          branch;
 int          ns1;
 int          ns2;
 int          result;
 filestr    * file;
 int          bits;
 int          fnext;


 err=NULL;

 ltool=(linktoolstr*)(toolbox->toolstate[toolbox->current].hprivate);
 branch=posn->view->branch;
 file=posn->file;
 result=1;
 bits=0;


 viewrefreshall(file,-1,VRTOOLS|VRONELAYER);
 linkdrag=0;

 if(findbox(posn,-1))
 {
  if(posn->object)
  {
   if(posn->otype!=OTEXTBOX) err=geterror(LI03);
   else
   if(posn->section!=start.section) err=geterror(LI04);
   else
   if(!branch && posn->object==start.object) err=geterror(LI06);
   else
   if(embededobject(posn->object,file)) err=geterror(LI13);
   else
   {
    textp=(textboxstr*)fob_find(posn->object,file);

    if(!branch && textp->master) err=geterror(LI14);
    else
    if(textp->data.repeat) err=geterror(LI02);
    else
    if(textp->fchain.prev) err=geterror(LI07);
    else
    if(childobject((chainobject*)textp,file)) err=geterror(LI15);
    else
    {
     texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
     src2=texsp->source;

     textp=(textboxstr*)fob_find(start.object,file);
     texsp=(texsrc*)(((char*)textp)+sizeof(textboxstr));
     src1=texsp->source;
     fnext=textp->fchain.next;

     if(!branch && src1==src2) err=geterror(LI08);
     else
     {
      ns1=nullsource(src1,file);
      ns2=nullsource(src2,file);
      if(!ns1 && !ns2) err=geterror(LI09);
      else
      {
       if(branch && (posn->page!=start.page || src1==src2))
       {
        if(!ns1 || !ns2)   err=geterror(LI05);
        else
        if(fnext) err=geterror(LI16);
        else
        if(posn->page!=start.page && 
          ((pagestr*)fob_find(start.page,file))->chain.prev)err=geterror(LI17);
        else
        {
         confirmparent(posn->view->handle,CONYN,&result,"{LI18}");
         if(src1==src2) bits|=LINKBNOLINK;
         bits|=LINKBCLOSE;
        }
       }
      }
     }

     if(!err && result>0)
     {
      err=linkbox2box(start.object,posn->object,posn->section,file,bits);
     }
    }
   }
  }
 }

 return(err);

 USE(box);
}


static os_error * linkmove(toolboxstr * toolbox,posnstr * posn,int move)
{
 ltool=(linktoolstr*)(toolbox->toolstate[toolbox->current].hprivate);

 if(move && linkdrag)
 {
  viewrefreshall(posn->file,-1,VRTOOLS|VRONELAYER);
  end=*posn;
  viewrefreshall(posn->file,-1,VRTOOLS|VRONELAYER);

  toolschangepointer(0,USERPOINTER,&linkpointer);
 }

 return(NULL);
}



static os_error * liprefresh(viewstr * view,int handle,int code)
{
 if(code==-1) 
 {
  writeiconf(view->info,IP_HELPI,"%s","{LI10}");
 }

 return(NULL);
 USE(handle);
 USE(view);
}




static os_error * lipclick(int handle,int userhandle,wimp_mousestr * m)
{
 return((m->i==LINKDELICON)?cutlink():NULL);
 USE(handle);
 USE(userhandle);
}





static ipstr lipstr=
{
 lipclick,       /* click    */
 NULL,           /* key      */
 lipmap,         /* icon map */
 TLINKTOOL,      /* template */
 0,              /* code     */
 liprefresh,     /* refresh  */

};


static os_error * linkselect(toolboxstr * toolbox,viewstr * view,
                                           filestr * file,int select)
{
 ltool=(linktoolstr*)(toolbox->toolstate[toolbox->current].hprivate);

 ltool->linkmode=select;

 if(!undoactive(file))
 {
  if(select)
  {
   clearselectionlo();
  }
  else
  {
   slink(file,0,ltool->linksection);
  }
 }


 if(select)
 {
  viewrefreshall(file,-1,VRTOOLS);
 }
 else
 {
  viewrefreshall(file,-1,VRBACK);
 }


 ltool->linkbox=0;

 ipchangepane(toolbox,IP_PANE1,select?&lipstr:NULL);

 return(NULL);

 USE(view);
}


static os_error * linkptr(toolboxstr * toolbox,posnstr * posn)
{
 toolschangepointer(0,USERPOINTER,&linkpointer);

 return(NULL);

 USE(posn);
 USE(toolbox);
}


static os_error * linkcancel(toolboxstr * toolbox,viewstr * view,
                                                  filestr * file)
{
 ltool=(linktoolstr*)(toolbox->toolstate[toolbox->current].hprivate);

 if(linkdrag)
 {
  viewrefreshall(file,-1,VRTOOLS|VRONELAYER);
  linkdrag=0;
  toolsstopdrag();
 }

 return(NULL);
 USE(view);
}


static os_error * linkmenu(toolboxstr * toolbox,viewstr * view,
                                         filestr * file,wimp_mousestr * m)
{
 return(openmenu("main_menu"));
 USE(view);
 USE(file);
 USE(m);
 USE(toolbox);
}





/* called by keyboard code to cut current link */

static os_error * cutlink(void)
{
 os_error * err;

 if(ltool->linkbox) err=removelink(ltool->linkbox,ltool->linksection,menufile);
 else               err=NULL;

 return(err);
}



static os_error * linkkey(toolboxstr * toolbox,int * key,fkeyfn * fk)
{
 ltool=(linktoolstr*)(toolbox->toolstate[toolbox->current].hprivate);

 if(*key==DEL)
 {
  *key=-1;
  return(cutlink());
 }
 else
 {
  menukey(key);
  if(*key==-1) *fk=menukey;
  return(NULL);
 }
}




static os_error * lmess(toolboxstr * toolbox,filestr * file,
                                      int message,void * data)
{
 if(message==TSAVEPOSN) lsaveposn(file);

 return(NULL);

 USE(toolbox);
 USE(data);
}








static toolstr link=
{
 TL_LINK,
 "link",
 "t_link\0t_plink",
 "{LNKTIP}",
 1,
 linkclick,
 linkmove,
 linkselect,
 linkptr,
 linkcancel,
 linkredraw,
 linkdragend,
 linkmenu,
 linkkey,
 NULL,
 NULL,
 lmess,

 0,

 0,       /*  preselect */
 0,       /*  noicon    */
 0,       /*  spare2    */
 0,       /*  spare3    */
 0,       /*  spare4    */ 
 0,       /*  spare5    */
 0,       /*  spare6    */
 0,       /*  spare7    */

 TPOINTER,
 0,
 sizeof(linktoolstr),
 5,       /*  rank      */
};





static errtag errortable[]=
{
 LI00,"{LI00}",
 LI01,"{LI01}",
 LI02,"{LI02}",
 LI03,"{LI03}",
 LI04,"{LI04}",
 LI05,"{LI05}",
 LI06,"{LI06}",
 LI07,"{LI07}",
 LI08,"{LI08}",
 LI09,"{LI09}",
 LI11,"{LI11}",
 LI12,"{LI12}",
 LI13,"{LI13}",
 LI14,"{LI14}",
 LI15,"{LI15}",
 LI16,"{LI16}",
 LI17,"{LI17}",

 ELAST ,""
};

static errlink errorlink;




#define UNDOS 5

static undofnstr undos[UNDOS]=
{
 unlink,        NULL,          ULINK,        "{ULI0}",
 undelink,      undelinkscrap, UDELINK,      "{ULI1}",
 unslink,       NULL,          USLINK,       "{ULI2}",
 usopen,        NULL,          USOPEN,       "{ULI3}",
 usclose,       NULL,          USCLOSE,      "{ULI4}",
};




static os_error * lxplinks(int text,int * message,
                                        plinkdatastr * plink,filestr * file)
{
 int           page;
 pagestr     * pagep;
 textboxstr  * textp;

 if(text)
 {
  textp=(textboxstr*)fob_find(text,file);
  if(!textp->data.repeat)
  {
   page=pageof((chainobject*)textp,file);
   pagep=(pagestr*)fob_find(page,file);
   if(pagep->data.number<plink->nf || pagep->data.number>plink->nl) *message=0;
  }
 }

 return(NULL);
}


os_error * lxplink(textboxstr * textp,int * message,
                                          plinkdatastr * plink,filestr * file)
{
 int           next;
 int           prev;

 next=textp->fchain.next;
 prev=textp->fchain.prev;

 lxplinks(next,message,plink,file);
 lxplinks(prev,message,plink,file);

 return(NULL);
}



os_error * linkinit(void)
{
 os_error * err;

 err=toolregister(&link);
 if(!err) err=adderrortable(errortable,&errorlink);
 if(!err) err=undoaddblock(undos,UNDOS);

 return(err);
}

