/*->c.stylemap */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "wos.h"
#include "flex.h"
#include "font.h"
#include "transform.h"

#include "etc.h"
#include "err.h"
#include "crc.h"
#include "xext.h"
#include "bits.h"
#include "xmath.h"
#include "fsx.h"
#include "bf.h"



#include "str.h"
#include "constants.h"

#include "reslink.h"
#include "xredraw.h"
#include "object.h"

#include "xcolour.h"

#include "file.h"
#include "view.h"
#include "undo.h"


#include "txstr.h"
#include "tx.h"
#include "vl.h"
#include "text.h"
#include "codes.h"
#include "fox.h"
#include "tabs.h"
#include "style.h"
#include "ustyle.h"
#include "act.h"
#include "fmap.h"
#include "fl.h"
#include "txchoices.h"


#include "stylemap.h"





#define MAPCHUNK 64*sizeof(mapstr)

mapstr * chainmap;
int      chainentries;
mapstr * absmap;
int      absentries;
mapstr * usermap;
int      userentries;
mapstr * colourmap;
int      colourentries;
mapstr * tabmap;
int      tabentries;




/* look up old number in map, and return new number or 0 */

static int cmpmap(const void * p1,const void * p2)
{
 mapstr * q1;
 mapstr * q2;
 q1=(mapstr *)p1;
 q2=(mapstr *)p2;
 return(q1->old-q2->old);
}



static int mapfind(int old,mapstr * map,int max)
{
 mapstr * r;
 mapstr   maphdr;

 maphdr.old=old;

 r=(mapstr*)bsearch(&maphdr,map,max,sizeof(mapstr),cmpmap);
 if(r) return(r->xnew);
 else  return(0);
}



static os_error * mapadd(int old,int xnew,mapstr ** map,int * maxp)
{
 os_error * err;
 mapstr   * r;
 mapstr     maphdr;
 int        n;
 int        max;

 max=*maxp;

 err=flex_chunk((flex_ptr)map,sizeof(mapstr)*(max+1),MAPCHUNK);

 maphdr.old=old;

 r=(mapstr*)bnearest(&maphdr,*map,max,sizeof(mapstr),cmpmap);

 n=r-(*map);
 memmove((*map)+n+1,(*map)+n,sizeof(mapstr)*(max-n));
 (*map)[n].old=old;
 (*map)[n].xnew=xnew;

 (*maxp)++;

 return(err);
}




static void trashmaps(void)
{
 flex_free((flex_ptr)&chainmap);
 flex_free((flex_ptr)&absmap);                                                   flex_free((flex_ptr)&usermap);
 flex_free((flex_ptr)&colourmap);
 flex_free((flex_ptr)&tabmap);
}


static os_error * initmaps(void)
{
 os_error * err;

 chainentries=0;
 absentries=0;
 userentries=0;
 colourentries=0;
 tabentries=0;

          err=flex_alloc((flex_ptr)&chainmap,MAPCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&absmap,MAPCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&usermap,MAPCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&colourmap,MAPCHUNK);
 if(!err) err=flex_alloc((flex_ptr)&tabmap,MAPCHUNK);

 if(err) trashmaps();

 return(err);
}




/*

  if colour name exists in dest, then if same colour copy
                                      if different
                                         if rename, create new colour
                                         else       use dest colour

  if style name exists in dest, then if same style copy


*/




static os_error * docolourmap(filestr * sfile,filestr * dfile,
                                              colourvalue * cv,int mode)
{
 os_error * err;
 int        cnew;
 int        key;

 err=NULL;

 key=cv->data.tint.key;
 cnew=mapfind(key,colourmap,colourentries);
 if(!cnew)
 {
  err=mapcolour(sfile,dfile,cv,mode);
  mapadd(key,cv->data.tint.key,&colourmap,&colourentries);
 }
 else
 {
  cv->data.tint.key=cnew;
 }

 return(err);
}




static os_error * docolourmapstyledata(actionbits * action,datawords * data,
                                    filestr * sfile,filestr * dfile,int mode)
{
 if(!action || action->fcolour)  docolourmap(sfile,dfile,&data->fcolour,mode);
 if(!action || action->bcolour)  docolourmap(sfile,dfile,&data->bcolour,mode);
 if(!action || action->underline)docolourmap(sfile,dfile,&data->ulcolour,mode);
 if(!action || action->strikeout)docolourmap(sfile,dfile,&data->stcolour,mode);
 if(!action || action->rulevert) 
                           docolourmap(sfile,dfile,&data->vrulecolour,mode);
 if(!action || action->rulehoriz)
                           docolourmap(sfile,dfile,&data->hrulecolour,mode);
 return(NULL);
}



static os_error * adduserstyles(userstylestr * us,filestr * sfile,
                                filestr * dfile,int * cnew,int overwrite,int copyrank)
{
 os_error * err;
 int        tnew;
 int        wasnew;


 wasnew=tnew=mapfind(us->style.data.tab,tabmap,tabentries);
 if(!tnew)
 {
  gettabruler(sfile,us->style.data.tab);
  addtabruler(dfile,&tabruler,0,tabrulern,&tnew);
  mapadd(us->style.data.tab,tnew,&tabmap,&tabentries);
 }

 us->style.data.tab=tnew;

 docolourmapstyledata(NULL,&us->style.data,sfile,
                                           dfile,overwrite?LOCOVERWRITE:0);

 err=adduserstyle2(us,dfile,cnew,overwrite,copyrank);

 if(!wasnew && tnew) remtabruler(dfile,tnew);

 return(err);
}


/* passed pointer to user style */

static os_error * dostylemap(filestr * sfile,filestr * dfile,
                                      int * ustyle,int * change,int mode)
{
 os_error    * err;
 int           cnew;
 int           key;
 userstylestr  us;
 userstylestr  ud;
 int           n=0;
 int           bstyle;

 err=NULL;

 key=*ustyle;

 if(key!=BULLETNOSTYLE)
 {
  cnew=mapfind(key,usermap,userentries);
  if(!cnew)
  {
   err=finduserstyle(&us,key,sfile);
   if(!err)
   {
    if(us.style.action.bullet)
    {
     bstyle=us.style.data.bulletstyle;
     if(bstyle!=key)
     {
      err=dostylemap(sfile,dfile,&bstyle,change,mode);
      if(!err)
      {
       us.style.data.bulletstyle=bstyle;
      }
     }
     else
     {
      us.style.data.bulletstyle=BULLETSELFSTYLE;
     }
    }
    else
    {
     us.style.data.bulletstyle=BULLETNOSTYLE;
    }

    /* from here, bulletstyle is valid in the destination file */
    /* so comparisions and addstyles work */

    if(!err)
    {
     if(finduserstylebyname(us.name,dfile,&ud)>=0) cnew=ud.style.n;
     else                                          cnew=0;

     if(cnew)
     {
      if(cmpustylesf(&us,sfile,&ud,dfile))
      {
       if(mode & LOCOVERWRITE)
       {
        err=adduserstyles(&us,sfile,dfile,&cnew,1,0);
       }
       else
       if(txrename)
       {
        n=0;
        while(1)
        {
         nextstylename(us.name,&n);
         if(finduserstylebyname(us.name,dfile,&ud)>=0)
         {
          /* dprintf(0,"try %s",us.name); */

          if(!cmpustylesf(&us,sfile,&ud,dfile))
          {
           cnew=ud.style.n;
           break;
          }
          /* dprintf(0,"failed"); */
         }
         else
         {
          err=adduserstyles(&us,sfile,dfile,&cnew,0,0);
          break;
         }
        }
       }
      }
     }
     else
     {
      adduserstyles(&us,sfile,dfile,&cnew,0,0);
     }

     mapadd(key,cnew,&usermap,&userentries);
    }
   }
  }

  *ustyle=cnew;
 }

 return(err);
}



static os_error * txremapsources(int source,int start,int len,
                                     filestr * sfile,filestr * dfile,int mode)
{
 os_error  * err;
 sourcestr * sourcep;
 char      * p;
 int         c;
 int         bits;
 int         offset;
 int         end;
 int         sn;
 int         snew;
 absstylestr s;
 int         chain[MAXCHAIN];
 int         nchain;
 int         cnew;
 int         tnew;
 int         wastnew;
 int         wascnew;
 int         i;
 int         change;
 int         seenbody;
 int         bstyle;


 err=fob_lock((fobjptr)&sourcep,source,dfile);
 if(!err)
 {
  if(len<0) len=sourcep->len;
  offset=start;
  end=start+len;

  if(sfile==dfile)
  {
   while(offset<end)
   {
    p=txptr(sourcep,offset);
    c=*p;
    bits=codex[c];

    if(c==SNO)
    {
     sn=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
     dupabsstyle(dfile,sn);
    }
    offset+=bits & CODELEN;
   }
  }
  else
  {
   if(!err)
   {
    while(offset<end)
    {
     p=txptr(sourcep,offset);
     c=*p;
     bits=codex[c];

     if(c==SNO)
     {
      sn=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
      snew=mapfind(sn,absmap,absentries);
      if(snew)
      {
       dupabsstyle(dfile,snew);
      }
      else
      {
       findabsstyle(sfile,sn,&s);

       change=0;
       seenbody=0;
       wascnew=cnew=mapfind(s.chain,chainmap,chainentries);
       if(!cnew)
       {
        findstylechains(sfile,s.chain,chain,&nchain);
        for(i=0;i<nchain;i++)
        {
         dostylemap(sfile,dfile,&chain[i],&change,mode);
         if(chain[i]==BODYTEXT) seenbody=1;
        }
        if(!seenbody) chain[nchain++]=BODYTEXT;

        addstylechain(dfile,chain,nchain,&cnew,1);
        mapadd(s.chain,cnew,&chainmap,&chainentries);
       }
       s.chain=cnew;


       if(s.style.action.bullet)
       {
        bstyle=s.style.data.bulletstyle;
        dostylemap(sfile,dfile,&bstyle,&change,mode);
        s.style.data.bulletstyle=bstyle;
       }
       else
       {
        s.style.data.bulletstyle=BULLETNOSTYLE;
       }


       docolourmapstyledata(&s.style.action,&s.style.data,sfile,dfile,mode);

       wastnew=tnew=0;
       if(s.style.action.tab)
       {
        wastnew=tnew=mapfind(s.style.data.tab,tabmap,tabentries);
        if(!tnew)
        {
         gettabruler(sfile,s.style.data.tab);
         addtabruler(dfile,&tabruler,0,tabrulern,&tnew);
         mapadd(s.style.data.tab,tnew,&tabmap,&tabentries);
        }
        s.style.data.tab=tnew;
       }

       fillinstyle(&s,dfile);

       addabsstyle(dfile,&s,1);
       if(!wastnew && tnew) remtabruler(dfile,tnew);
       if(!wascnew && cnew) remstylechain(dfile,cnew,1);

       snew=s.style.n;
       mapadd(sn,snew,&absmap,&absentries);
      }

      p=txptr(sourcep,offset);
      *(p+1)=(char)snew;
      *(p+2)=(char)(snew>>8);
      *(p+3)=(char)(snew>>16);
     }
     offset+=bits & CODELEN;
    }
   }
  }

  fob_unlock((fobjptr)&sourcep,dfile);
 }
 return(err);
}



os_error * txremapsource(int source,int start,int len,
                                 filestr * sfile,filestr * dfile,int mode)
{
 os_error * err;

 err=NULL;

 if(sfile!=dfile) err=initmaps();
 if(!err)
 {
  err=txremapsources(source,start,len,sfile,dfile,mode);

  if(sfile!=dfile) trashmaps();
 }

 return(err);
}






os_error * txcopysourcedata(int s,int * d,int sv,int * dv,
                                filestr * sfile,filestr * dfile,int mode)
{
 os_error   * err;
 vsourcestr * vp;
 int          vloff;
 int          vmax;
 vlstr      * v;

 err=NULL;

 if(sfile!=dfile) err=initmaps();
 if(!err)
 {
           err=txsourcecopy(s,d,sfile,dfile);
  if(!err) err=txremapsources(*d,0,-1,sfile,dfile,mode);
  if(!err)
  {
   vsourcecopy(sv,dv,sfile,dfile);

   if(sfile!=dfile)
   {
    fob_mod(*dv,dfile,(fobjptr)&vp);

    vmax=vp->goffset;
    v=(vlstr*)vp->data;
    for(vloff=0;vloff<vmax;vloff+=sizeof(vlstr),v++)
    {
     if(!v->line.xl)
     {
      v->line.style=mapfind(v->line.style,absmap,absentries);
     }
    }

    vloff+=vp->maxlen-vp->len;
    v=(vlstr*)(((char*)v)+(vp->maxlen-vp->len)); 

    vmax=vp->maxlen;
    for(;vloff<vmax;vloff+=sizeof(vlstr),v++)
    {
     if(!v->line.xl)
     {
      v->line.style=mapfind(v->line.style,absmap,absentries);
     }
    }
   }
  }
  if(sfile!=dfile) trashmaps();
 }

 return(err);
}





os_error * txcopyeobs(filestr * sfile,filestr * dfile,transformstr * tr,
                      int dtext,int dpage,int dsection,int dsource,
                      int start,int len,int mode)
{
 os_error  * err;
 sourcestr * sourcep;
 char      * p;
 int         c;
 int         bits;
 int         offset;
 int         end;
 objectloc   sloc;
 objectloc   dloc;


 err=fob_lock((fobjptr)&sourcep,dsource,dfile);
 if(!err)
 {
  if(len<0) len=sourcep->len;
  offset=start;
  end=start+len;

  while(offset<end)
  {
   p=txptr(sourcep,offset);
   c=*p;
   bits=codex[c];

   if(c==EOB)
   {
    memset(&sloc,0,sizeof(objectloc));
    memset(&dloc,0,sizeof(objectloc));

    sloc.object=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16)+((*(p+4))<<24);

    sloc.file=sfile;

    dloc.file=dfile;
    dloc.page=dpage;
    dloc.section=dsection;
    dloc.parent=dtext;
    dloc.chain=-1;         /* no chain */
    dloc.copymode=LOCFLOWDATA | (mode & LOCOVERWRITE) | LOCNOFLOW;

    err=objectcopy(&sloc,&dloc,tr);

    p=txptr(sourcep,offset);
    *(p+1)=(char)dloc.object;
    *(p+2)=(char)(dloc.object>>8);
    *(p+3)=(char)(dloc.object>>16);
    *(p+4)=(char)(dloc.object>>24);

   }
   offset+=bits & CODELEN;
  }

  fob_unlock((fobjptr)&sourcep,dfile);
 }
 return(err);
}





os_error * copyuserstyles(filestr * sfile,filestr * dfile)
{
 os_error         * err;
 userstylestr       us;
 int                cnew;
 int                styles;
 userstylebasestr * sbase;
 int                i;

 err=initmaps();
 if(!err)
 {
  styles=sfile->hdr.ustyles;

  i=0;
  while(1)
  {
   sbase=(userstylebasestr*)fob_find(styles,sfile);
   if(i>=sbase->n) break;
   us=sbase->userstyle[i];
   adduserstyles(&us,sfile,dfile,&cnew,0,1);
   i++;
  }

  trashmaps();
 }

 return(err);
}


/****************************************************************************/

static os_error * projectcoloursubs(colourvalue * cv)
{
 int cnew;
 int key;

 key=cv->data.tint.key;

 cnew=mapfind(key,colourmap,colourentries);
 if(!cnew) mapadd(key,key,&colourmap,&colourentries);

 return(NULL);
}


static os_error * projectcolour(datawords * data)
{
 projectcoloursubs(&data->fcolour);
 projectcoloursubs(&data->bcolour);
 projectcoloursubs(&data->ulcolour);
 projectcoloursubs(&data->stcolour);
 projectcoloursubs(&data->vrulecolour);
 return(projectcoloursubs(&data->hrulecolour));
}


static os_error * projectfont(int mfont)
{
 int cnew;

 cnew=mapfind(mfont,tabmap,tabentries);
 if(!cnew)
 {
  mapadd(mfont,mfont,&tabmap,&tabentries);
 }

 return(NULL);
}




static os_error * txprojectuser(int un,filestr * file)
{
 int           cnew;
 userstylestr  us;


 cnew=mapfind(un,usermap,userentries);
 if(!cnew)
 {
  finduserstyle(&us,un,file);
  if(us.style.n)
  {
   projectcolour(&us.style.data);
   projectfont(us.style.data.mfont);

   mapadd(un,un,&usermap,&userentries);
   txprojectuser(us.style.data.bulletstyle,file);
  }
 }

 return(NULL);
}




static os_error * txproject(int sn,filestr * file)
{
 absstylestr   s;
 int           chain[MAXCHAIN];
 int           snew;
 int           cnew;
 int           nchain;
 int           i;


 if(sn)
 {
  snew=mapfind(sn,absmap,absentries);
  if(!snew)
  {
   findabsstyle(file,sn,&s);

   cnew=mapfind(s.chain,chainmap,chainentries);
   if(!cnew)
   {
    findstylechains(file,s.chain,chain,&nchain);
    for(i=0;i<nchain;i++)
    {
     txprojectuser(chain[i],file);
    }
    mapadd(s.chain,s.chain,&chainmap,&chainentries);
   }

   projectcolour(&s.style.data);
   projectfont(s.style.data.mfont);
   mapadd(sn,sn,&absmap,&absentries);
   txprojectuser(s.style.data.bulletstyle,file);
  }
 }

 return(NULL);
}




os_error * txprojectstart(int source,int sn,int offset,int maxoffset,
                                                           filestr * file)
{
 os_error  * err;
 sourcestr * sourcep;
 int         bits;
 int         slen;
 int         c;
 char      * p;
 int         coffset;
 int         snoffset;


 err=initmaps();
 if(!err)
 {
  err=fob_lock((fobjptr)&sourcep,source,file);
  if(!err)
  {
   if(maxoffset<0) maxoffset=sourcep->len;

   coffset=0;
   snoffset=-1;

   /* code to ensure a style sequence always starts text */

   while(1)
   {
    p=txptr(sourcep,coffset);
    c=*p;
    bits=codex[c];
    slen=bits & CODELEN;
    if(c==SNO) snoffset=coffset;
    if(coffset>=offset) break;
    coffset+=slen;
    if(slen<=0) break;
   }

   if(snoffset>=0 && snoffset!=offset)
   {
    p=txptr(sourcep,snoffset);
    sn=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
    txproject(sn,file);
   }


   do
   {                                                                                p=txptr(sourcep,offset);
    c=*p;
    bits=codex[c];
    slen=bits & CODELEN;

    if(c==SNO)
    {
     sn=(*(p+1))+((*(p+2))<<8)+((*(p+3))<<16);
     txproject(sn,file);
    }

    offset+=slen;

    if(offset>=maxoffset) break;
   } while(c!=EOT);

   fob_unlock((fobjptr)&sourcep,file);
  }

  if(err) trashmaps();
 }

 return(err);
}



os_error * txprojectend(void)
{
 trashmaps();
 return(NULL);
}


