/*->c.vl */


#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>  
#include <ctype.h>
#include <assert.h>


#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "deb.h"

#include "transform.h"

#include "etc.h"
#include "err.h"
#include "wos.h"


#include "str.h"
#include "constants.h"

#include "file.h"

#include "txstr.h"
#include "tx.h"
#include "vl.h"
#include "fl.h"


/*                                                                                  handles buffers of vlines at a low level 
*/



#define VLCHUNK 0x40*sizeof(vlstr)


os_error * vsourcenew(int * key,filestr * file)
{
 os_error   * err;
 vsourcestr * vsourcep;

 err=fob_create(sizeof(vsourcestr)+VLCHUNK,file,OVSOURCE,(fobjptr)&vsourcep);
 if(!err)
 {
  vsourcep->len=0;
  vsourcep->maxlen=0;
  vsourcep->goffset=0;
  *key=vsourcep->key;
 }
 return(err);
}


os_error * vsourcezero(int vsrc,filestr * file)
{
 os_error   * err;
 vsourcestr * vsourcep;

 err=fob_mod(vsrc,file,(fobjptr)&vsourcep);

 if(!err)
 {
  vsourcep->len=0;
  vsourcep->maxlen=0;
  vsourcep->goffset=0;
 }

 return(err);
}


os_error * vsourcecopy(int src,int * key,filestr * sfile,filestr * dfile)
{
 os_error  * err;
 vsourcestr * vsourcep;


 err=fob_copy((fobjptr)&vsourcep,dfile,src,sfile);

 *key=vsourcep->key;
 vsourcep->fldstack=0;    /* either copy it or get rid of it */


 return(err);
}




os_error * vsourcedel(int key,filestr * file)
{
 if(key)
 {
  flowstackunlink(key,file);
  return(fob_delete(key,file));
 }
 else
	{
  return(NULL);
	}
}



static void setgapto(vsourcestr * source,int offset)
{
 char * buff;
 int    len;
 int    maxlen;
 int    goffset;
 int    gapsize;

 buff=source->data;
 len=source->len;
 maxlen=source->maxlen;
 goffset=source->goffset;
 gapsize=maxlen-len;

 if(offset>goffset) offset+=gapsize;

 if(goffset>offset)
 {
  memmove(buff+offset+gapsize,buff+offset,goffset-offset);
  source->goffset=offset;
 }
 else
 if(offset>goffset)
 {
  memmove(buff+goffset,buff+goffset+gapsize,offset-goffset-gapsize);
  source->goffset+=offset-goffset-gapsize;
 }
}



os_error * vsourceinsert(vlstr * v,int n,textsrcstr * textsrc,filestr * file)
{
 os_error   * err;
 int          free;
 int          xnew;
 vsourcestr * vsourcep;
 int          topgapoffset;

 err=NULL;
 setgapto(textsrc->vsourcep,textsrc->voffset);
 free=textsrc->vsourcep->maxlen-textsrc->vsourcep->len;

 if(free<n)
 {

  xnew=n-free+VLCHUNK;

/*  cprintf("1new=%d n=%d free=%d maxlen=%d set=%d",xnew,n,free,
                textsrc->vsourcep->maxlen,
                sizeof(vsourcestr)+textsrc->vsourcep->maxlen+xnew);
*/

  err=fob_change(textsrc->vsource,file,
                      sizeof(vsourcestr)+textsrc->vsourcep->maxlen+xnew,
                                                      (fobjptr)&vsourcep);


  if(!err)
  {
   topgapoffset=textsrc->voffset+free;

   memmove(vsourcep->data+topgapoffset+xnew,
           vsourcep->data+topgapoffset,
           vsourcep->maxlen-topgapoffset);

   vsourcep->maxlen+=xnew;
   free=vsourcep->maxlen-vsourcep->len;
  }
 }

 if(!err)
 {
  textsrc->vsourcep->len+=n;
  memmove(textsrc->vsourcep->data+textsrc->voffset,v,n);
  textsrc->vsourcep->goffset+=n;
 }

 return(err);
}


os_error * vsourceclose(textsrcstr * textsrc,filestr * file)
{
 os_error   * err;
 int          free;
 vsourcestr * vsourcep;
 int          len;

 err=NULL;

 len=textsrc->vsourcep->len;

 free=textsrc->vsourcep->maxlen-len;

 if(free)
 {
  setgapto(textsrc->vsourcep,len);

  err=fob_change(textsrc->vsource,file,
                   sizeof(vsourcestr)+len,(fobjptr)&vsourcep);
  textsrc->vsourcep->maxlen=len;
 }

 return(err);
}



os_error * vsourceremove(int n,textsrcstr * textsrc,filestr * file)
{
 if(n>0 && textsrc->voffset<textsrc->vsourcep->len)
 {
  setgapto(textsrc->vsourcep,textsrc->voffset);
  textsrc->vsourcep->len-=n;
 }
 return(NULL);
 USE(file);
}


os_error * getvline(vlstr * vline,vsourcestr * v,int vloff)
{

#ifdef _DEBUG
 assert(vloff<v->len && vloff>=0);
 assert(v->goffset<=v->len && v->goffset>=0);
#else


#endif


 if(vloff>=v->goffset) vloff+=v->maxlen-v->len;
 *vline=*(vlstr*)(v->data+vloff);
 return(NULL);
}


vlstr * vlptr(vsourcestr * v,int vloff)
{
#ifdef _DEBUG
 assert(vloff<v->len && vloff>=0);
 assert(v->goffset<=v->len && v->goffset>=0);
#else
	if(!(vloff<v->len && vloff>=0)) return(NULL);
 if(!(v->goffset<=v->len && v->goffset>=0)) return(NULL);
#endif

 if(vloff>=v->goffset) vloff+=v->maxlen-v->len;
 return((vlstr*)(v->data+vloff));
}



int vsourcesize(int src,filestr * file)
{
 vsourcestr * v;

 v=(vsourcestr*)fob_find(src,file);

 return(v->len);
}



os_error * getlinearea(vector * s,aboxstr * lbox,aboxstr * zbox,aboxstr * mbox)
{
 vlstr   vline;
 aboxstr abox;

 getvline(&vline,s->textsrc.vsourcep,s->textsrc.voffset);

 abox.x0=vline.line.x;
 abox.x1=abox.x0+MAXWX(vline);
 abox.y0=LY(vline.line.yy);
 abox.y1=abox.y0+vline.line.h;

 if(zbox) *zbox=abox;

 abox.x1=abox.x0+vline.line.w;
 if(lbox) *lbox=abox;
 if(mbox) *mbox=abox;

 return(NULL);
}


