/*->c.timex */

#include <stdafx.h>


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>


#include "os.h"
#include "fsx.h"
#include "wimpt.h"

#include "timex.h"
#include "bits.h"
#include "deb.h"
#include "etc.h"

#include "ostime.h"


/*****************************************************************************/

// fill in later to handle wrap of ticker time

int tickdiff(int t1,int t2)
{
 return(t1-t2);
}


int monotonic(void)
{
 return(GetTickCount()/10);
}



/* get current time */

void getcurtime(arctime * time)
{
 FILETIME ftime;

 GetSystemTimeAsFileTime(&ftime);
 osfiletime2riscostime(time,&ftime);
}



int filetime(fstat * f,arctime * t)
{
 if((f->load & 0xFFF00000)==0xFFF00000)   /* has time stamp */
 {
  t->lo=f->exec;
  t->hi=f->load & 0xFF;
  return(1);
 }
 else
  return(0);
}


void setfiletime(fstat * f,arctime * t)
{
 f->load&=~0xFF;
 f->load|=(0xFFF00000|t->hi);
 f->exec=t->lo;
}


/* return -1 f1<f2 1 if f1>f2 else 0 */

int cmptimes(fstat * f1,fstat * f2)
{
 arctime t1;
 arctime t2;
 int     t1true;
 int     t2true;

 t1true=filetime(f1,&t1);
 t2true=filetime(f2,&t2);

 if(!t1true && !t2true) return(0);
 else
 if(!t1true)            return(-1);
 else
 if(!t2true)            return(1);

 if(t1.hi==t2.hi)
 {
  if(t1.lo>t2.lo) return(1);
  else            return(-1);
 }
 else
 if(t1.hi>t2.hi) return(1);
 else            return(-1);
}



/* return -1 t1<t2 1 if t1>t2 else 0 */

int cmptimex(arctime * t1,arctime * t2)
{

 if(t1->hi==t2->hi)
 {
  if(t1->lo>t2->lo) return(1);
  else              return(-1);
 }
 else
 if(t1->hi>t2->hi) return(1);
 else              return(-1);
}


void addtimex(arctime * t,unsigned int delta,int inc)
{
 unsigned int oldlo;

 oldlo=t->lo;

 if(inc)
 {
  t->lo+=delta;
  if(t->lo<oldlo) t->hi++;
 }
 else
 {
  t->lo-=delta;
  if(t->lo>oldlo) t->hi--;
 }
}



/*

Code	Meaning	Example
CS	centiseconds	99
SE	seconds	59
MI	minutes	45
12	hours in 12-hour format	07
24	hours in 24-hour format	19
AM	am or pm	am
PM	am or pm	pm
WE	weekday, in full	Friday
W3	weekday, first 3 letters only	Thu
WN	weekday, as a number	5
DY	day of month	12
ST	suffix st, nd, rd or th	st
MO	month, in full	March
M3	month, first 3 letters only	Jan
MN	month, as a number	12
CE	century	20
YR	year within century	95
WK	week of the year	52


*/

#define STMAX 31

static char * st_eng[STMAX+1]={"xx","st","nd","rd","th","th","th","th","th","th",
                               "th","th","th","th","th","th","th","th","th","th",
																															"th","st","nd","rd","th","th","th","th","th","th",
																															"th","st"
                              };


void writets(char * string,char * format,arctime * timex)
{
 os_error * err;
 FILETIME   filet;
 FILETIME   localt;
 SYSTEMTIME systemt;
 char       newformat[SYSDTLEN];
 struct tm  tmt;
 char     * r;
 char     * w;
 int        zeroflag;
 int        ch;

 err=NULL;


 riscostime2osfiletime(&filet,timex);
 FileTimeToLocalFileTime(&filet,&localt);
 FileTimeToSystemTime(&localt,&systemt);

 tmt.tm_hour=systemt.wHour;
 tmt.tm_isdst=0;
 tmt.tm_mday=systemt.wDay;
 tmt.tm_min=systemt.wMinute;
 tmt.tm_mon=systemt.wMonth-1;
 tmt.tm_sec=systemt.wSecond;
 tmt.tm_wday=systemt.wDayOfWeek;
 tmt.tm_year=systemt.wYear-1900;


 *string=0;

 r=format;
 w=newformat;
 while(1)
 {
  ch=*r++;

  if(ch=='%')
  {
   if(*r=='Z')
   {
    zeroflag=1;
    r++;
   }
   else
   {
    zeroflag=0;
   }

   if(cstrncmp(r,"0",1)==0)
   {
    *w++=0;
    r+=1;
   }
   else
   if(cstrncmp(r,"CS",2)==0)  // centiseconds
   {
    if(zeroflag) w+=sprintf(w,"%d",(timex->lo % 100));
    else         w+=sprintf(w,"%02d",(timex->lo % 100));
    r+=2;
   }
   else
   if(cstrncmp(r,"SE",2)==0)
   {
    if(zeroflag)
    {
     strcpy(w,"%#S");
     w+=3;
    }
    else
    {
     strcpy(w,"%S");
     w+=2;
    }
    r+=2;
   }
   else
   if(cstrncmp(r,"MI",2)==0)
   {
    if(zeroflag)
    {
     strcpy(w,"%#M");
     w+=3;
    }
    else
    {
     strcpy(w,"%M");
     w+=2;
    }
    r+=2;
   }
   else
   if(cstrncmp(r,"12",2)==0)
   {
    if(zeroflag)
    {
     strcpy(w,"%#I");
     w+=3;
    }
    else
    {
     strcpy(w,"%I");
     w+=2;
    }
    r+=2;
   }
   else
   if(cstrncmp(r,"24",2)==0)
   {
    if(zeroflag)
    {
     strcpy(w,"%#H");
     w+=3;
    }
    else
    {
     strcpy(w,"%H");
     w+=2;
    }
    r+=2;
   }
   else
   if(cstrncmp(r,"AM",2)==0)
   {
    strcpy(w,"%p");   // seemingly result depends on current locale settings
    w+=2;
    r+=2;
   }
   else
   if(cstrncmp(r,"PM",2)==0)
   {
    strcpy(w,"%p");   // seemingly result depends on current locale settings
    w+=2;
    r+=2;
   }
   else
   if(cstrncmp(r,"W3",2)==0) // three letter weekday
   {
    strcpy(w,"%a");
    w+=2;
    r+=2;
   }
   else
   if(cstrncmp(r,"WE",2)==0) // day in full
   {
    strcpy(w,"%A");
    w+=2;
    r+=2;
   }
   else
   if(cstrncmp(r,"WN",2)==0)  // weekday as number
   {
    strcpy(w,"%a");
    w+=2;
    r+=2;
   }
   else
   if(cstrncmp(r,"DY",2)==0)
   {
    if(zeroflag) w+=sprintf(w,"%d",systemt.wDay); // day of month
    else         w+=sprintf(w,"%02d",systemt.wDay);
    r+=2;

    if(cstrncmp(r,"%ST",3)==0)       // 1st, 2nd, 3rd, 4th
    {
					w+=sprintf(w,"%s",st_eng[MIN(systemt.wDay,STMAX)]);
     r+=3;
    }
			}
			else
   if(cstrncmp(r,"ST",2)==0)       // 1st, 2nd, 3rd, 4th
   {
 			w+=sprintf(w,"%s",st_eng[MIN(systemt.wDay,STMAX)]);
    r+=2;
   }
   else
   if(cstrncmp(r,"MO",2)==0)
   {
    strcpy(w,"%B");
    w+=2;
    r+=2;
   }
   else
   if(cstrncmp(r,"M3",2)==0)
   {
    strcpy(w,"%b");
    w+=2;
    r+=2;
   }
   else
   if(cstrncmp(r,"MN",2)==0)
   {
    if(zeroflag) w+=sprintf(w,"%d",systemt.wMonth);
    else         w+=sprintf(w,"%02d",systemt.wMonth);
    r+=2;
   }
   else
   if(cstrncmp(r,"CE%YR",5)==0)
   {
    strcpy(w,"%Y");
    w+=2;
    r+=5;
   }
   else
   if(cstrncmp(r,"CE",2)==0)
   {
    if(zeroflag) w+=sprintf(w,"%d",systemt.wYear/100);
    else         w+=sprintf(w,"%02d",systemt.wYear/100);
    r+=2;
   }
   else
   if(cstrncmp(r,"YR",2)==0)
   {
    if(zeroflag)
    {
     strcpy(w,"%#y");
     w+=3;
    }
    else
    {
     strcpy(w,"%y");
     w+=2;
    }
    r+=2;
   }
   else
   if(cstrncmp(r,"WK",2)==0)
   {
    if(zeroflag)
    {
     strcpy(w,"%#U");
     w+=3;
    }
    else
    {
     strcpy(w,"%U");
     w+=2;
    }
    r+=2;
   }
   else
   if(ch=='%')
   {
    *w++='%';
    *w++='%';
    r+=1;
   }
  }
  else
  {
   *w++=(char)ch;
  }

  if(ch==0) break;
 }


// cprintf("new %s old %s",newformat,format);

 strftime(string,SYSDTLEN,newformat,&tmt);

// cprintf("string %s old %s new %s",string,format,newformat);
}



void writedate(char * string,arctime * timex)
{
 writets(string,"%ZDY %M3 %YR%0",timex);
}


void writetime(char * string,arctime * timex)
{
 writets(string,"%Z12:%MI%AM%0",timex);
}



void writetimedate(char * string,arctime * timex)
{
 writets(string,"%24:%MI:%SE %W3 %DY %M3 %YR%0",timex);
}



char sysdstring[SYSDTLEN];
char syststring[SYSDTLEN];


void getsysdate(char * format,int d,int t)
{
 if(d && t) sprintf(format,"%s %s%%0",syststring,sysdstring);
 else
 if(t)      sprintf(format,"%s%%0",syststring);
 else       sprintf(format,"%s%%0",sysdstring);
}


void timexinitsys(char * d,char * t)
{

 if(d) strcpy(sysdstring,d);
 else
 {
  strcpy(sysdstring,"%ZDY %M3 %YR");
//  sprintf(progname,"%s$DateFormat",wimpt_programname());
//  if((p=getenv(progname))!=NULL) strcpy(sysdstring,p);
 }

 if(t) strcpy(syststring,t);
 else
 {
  strcpy(syststring,"%Z12:%MI%AM");
//  sprintf(progname,"%s$TimeFormat",wimpt_programname());
//  if((p=getenv(progname))!=NULL) strcpy(syststring,p);
 }
}




void writesysst(char * string,int d,int t)
{
 arctime timex;
 char format[SYSDTLEN];

 getcurtime(&timex);
 getsysdate(format,d,t);
 writets(string,format,&timex);
}



void writesystime(char * string)
{
 writesysst(string,0,1);
}



void writesysdate(char * string)
{
 writesysst(string,1,0);
}


void writesystimedate(char * string)
{
 writesysst(string,1,1);
}


void writefiletime(char * string,fstat * stat)
{
 arctime t;

 if(filetime(stat,&t))
   writetimedate(string,&t);
 else
   strcpy(string,"");
}


void writehms(char * str,int time)
{
 time=time/100;
 sprintf(str,"%02d:%02d:%02d",(time/3600) % 24,(time/60) % 60,time % 60);
}





int timebomb(int year,int month,int day)
{
 SYSTEMTIME systemt;

 GetSystemTime(&systemt);

 return(systemt.wYear>year || 
        (systemt.wYear==year && systemt.wMonth>month) ||
        (systemt.wYear==year && systemt.wMonth==month && systemt.wDay>day));
}
